from jammy.utils import num_check as check
import torch

__all__ = [
    "tile",
    "sum_except_batch",
    "split_leading_dim",
    "merge_leading_dims",
    "repeat_rows",
    "logabsdet",
    "random_orthogonal",
    "searchsorted",
    "cbrt",
    "get_temperature"
]

def tile(x, n):
    if not check.is_positive_int(n):
        raise TypeError("Argument 'n' must be a positive integer.")
    x_ = x.reshape(-1)
    x_ = x_.repeat(n)
    x_ = x_.reshape(n, -1)
    x_ = x_.transpose(1, 0)
    x_ = x_.reshape(-1)
    return x_


def sum_except_batch(x, num_batch_dims=1):
    """Sums all elements of `x` except for the first `num_batch_dims` dimensions."""
    if not check.is_nonnegative_int(num_batch_dims):
        raise TypeError("Number of batch dimensions must be a non-negative integer.")
    reduce_dims = list(range(num_batch_dims, x.ndimension()))
    return torch.sum(x, dim=reduce_dims)


def split_leading_dim(x, shape):
    """Reshapes the leading dim of `x` to have the given shape."""
    new_shape = torch.Size(shape) + x.shape[1:]
    return torch.reshape(x, new_shape)


def merge_leading_dims(x, num_dims):
    """Reshapes the tensor `x` such that the first `num_dims` dimensions are merged to one."""
    if not check.is_positive_int(num_dims):
        raise TypeError("Number of leading dims must be a positive integer.")
    if num_dims > x.dim():
        raise ValueError(
            "Number of leading dims can't be greater than total number of dims."
        )
    new_shape = torch.Size([-1]) + x.shape[num_dims:]
    return torch.reshape(x, new_shape)


def repeat_rows(x, num_reps):
    """Each row of tensor `x` is repeated `num_reps` times along leading dimension."""
    if not check.is_positive_int(num_reps):
        raise TypeError("Number of repetitions must be a positive integer.")
    shape = x.shape
    x = x.unsqueeze(1)
    x = x.expand(shape[0], num_reps, *shape[1:])
    return merge_leading_dims(x, num_dims=2)


def logabsdet(x):
    """Returns the log absolute determinant of square matrix x."""
    # Note: torch.logdet() only works for positive determinant.
    _, res = torch.slogdet(x)
    return res


def random_orthogonal(size):
    """
    Returns a random orthogonal matrix as a 2-dim tensor of shape [size, size].
    """

    # Use the QR decomposition of a random Gaussian matrix.
    x = torch.randn(size, size)
    q, _ = torch.qr(x)
    return q

def searchsorted(bin_locations, inputs, eps=1e-6):
    bin_locations[..., -1] += eps
    return torch.sum(inputs[..., None] >= bin_locations, dim=-1) - 1


def cbrt(x):
    """Cube root. Equivalent to torch.pow(x, 1/3), but numerically stable."""
    return torch.sign(x) * torch.exp(torch.log(torch.abs(x)) / 3.0)


def get_temperature(max_value, bound=1 - 1e-3):
    """
    For a dataset with max value 'max_value', returns the temperature such that

        sigmoid(temperature * max_value) = bound.

    If temperature is greater than 1, returns 1.

    :param max_value:
    :param bound:
    :return:
    """
    max_value = torch.Tensor([max_value])
    bound = torch.Tensor([bound])
    temperature = min(-(1 / max_value) * (torch.log1p(-bound) - torch.log(bound)), 1)
    return temperature


# This is an exmaple to take advantage of trainer
# The design of trainer is a simple model for quick prototype
# For training process, user needs to implement the loss_fn, and other visualizization functions
from jamtorch.trainer import CfgTrainer, hydpath
import numpy as np
import torch
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import Dataset, DataLoader
from jammy.image import imread, imwrite
from jammy.logging import Wandb
import jamtorch.prototype as jampt
from torch.optim.lr_scheduler import StepLR
import jamtorch.nn as nn
from omegaconf import OmegaConf
from einops import repeat
import hydra


def init_model(cfg):
    INPUT = 2 * cfg.fourier_dim if cfg.is_fourier else 2
    model = nn.MLP(INPUT, 3, 64, 3, is_batchnorm=cfg.is_batchnorm)
    return model


def mse_loss_fn(model, feed_dict, is_train):
    data, target = feed_dict
    data, target = data.float().to(jampt.device), target.to(jampt.device).float()
    output = model(data)
    loss = F.l1_loss(target, output)
    return loss, {}, {}

class ImgDataset(Dataset):
    def __init__(self, cfg):
        self.img = torch.FloatTensor(imread(hydpath(cfg.img)).copy()) / 255
        self.row = self.img.shape[0]
        self.column = self.img.shape[1]

        rows, columns = np.meshgrid(np.arange(self.column), np.arange(self.row))
        self.mesh_xs = torch.FloatTensor(
            np.stack([columns, rows], axis=2).reshape(-1, 2)
        ) # (row_th, col_th)
        if cfg.is_fourier:
            position = repeat(self.mesh_xs, 'w h -> w (c h)', c=cfg.fourier_dim)
            fre = np.arange(1, cfg.fourier_dim+1)[:,None] * np.array([2*np.pi / self.row, 2*np.pi / self.column])[None, :]
            fre = fre.reshape(1, -1) * 0.26
            self.mesh_xs = torch.sin(position * fre)

        self.pixels = self.img.view(self.column * self.row, 3)
        save_img = (self.pixels *255).view(self.row, self.column, 3)
        imwrite("gt.png", jampt.get_numpy(save_img).astype(np.uint8))

    def __len__(self):
        return self.row * self.column

    def __getitem__(self, idx):
        return self.mesh_xs[idx], self.pixels[idx]

def init_data(cfg):
    train_kwargs = {"batch_size": cfg.batch_size}
    test_kwargs = {"batch_size": cfg.batch_size}
    cuda_kwargs = {"shuffle": True, "pin_memory": True}
    trainset = ImgDataset(cfg)
    valset = ImgDataset(cfg)
    trainloader = DataLoader(trainset, **train_kwargs, **cuda_kwargs)
    valloader = DataLoader(valset, **test_kwargs, **cuda_kwargs)
    return trainloader, valloader

def eval_after_wraper(trainloader):
    trainset = trainloader.dataset
    def mse_val_after(trainer, *args):
        with torch.no_grad():
            output = trainer.ema.model(trainset.mesh_xs.float().to(jampt.device))
            img = output.view(trainset.row, trainset.column, 3)
            img = jampt.get_numpy(img)
            save_img = (img * 255).astype(np.uint8)
            imwrite(f"mse_{trainer.epoch_cnt}.jpg", save_img)
    return mse_val_after

def run(cfg):
    jampt.set_gpu_mode(cfg.cuda)
    trainloader, valloader = init_data(cfg.data)
    eval_after_fn = eval_after_wraper(trainloader)
    model = init_model(cfg.model).to(jampt.device)
    optimizer = optim.Adam(model.parameters(), lr=cfg.lr)
    scheduler = StepLR(optimizer, step_size=1, gamma=cfg.gamma)
    trainer = CfgTrainer(model, optimizer, mse_loss_fn, scheduler, cfg.trainer)
    trainer.register_event("val:after", eval_after_fn)
    trainer.set_monitor(cfg.trainer.wandb.log)
    trainer(trainloader, valloader)

@hydra.main(config_name="pixels.yaml")
def main(cfg):
    Wandb.launch(cfg, cfg.trainer.wandb.log, True)
    OmegaConf.set_struct(cfg, False)
    run(cfg)
    Wandb.finish()


if __name__ == "__main__":
    main()



import hashlib
import pickle

__all__ = ["md5_encodef", "md5_encode_obj"]

def md5_encodef(f_input):
    file_code = open(f_input, "rb").read()
    md5_hash = hashlib.md5()
    md5_hash.update(file_code)
    digest = md5_hash.hexdigest()
    return digest

def md5_encode_obj(f_obj):
    byte_obj = pickle.dumps(f_obj)
    md5_hash = hashlib.md5()
    md5_hash.update(byte_obj)
    digest = md5_hash.hexdigest()
    return digest



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from pallets_sphinx_themes import ProjectLink

# -- Project information -----------------------------------------------------

project = "Jam"
copyright = "2021, Qsh.27"
author = "Qsh.27"


# -- General configuration ---------------------------------------------------

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "pallets_sphinx_themes",
    "sphinx_tabs.tabs",
    "m2r2",
]
source_suffix = [".rst", ".md"]
autodoc_typehints = "description"
intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "click"
html_theme_options = {"index_sidebar_logo": False}
html_context = {
    "project_links": [
        ProjectLink("PyPI Releases", "https://pypi.org/project/jammy/"),
        ProjectLink("Source Code", "https://gitlab.com/zqsh419/jam"),
    ]
}
html_sidebars = {
    "index": ["project.html", "localtoc.html", "searchbox.html", "ethicalads.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html", "ethicalads.html"],
}
singlehtml_sidebars = {"index": ["project.html", "localtoc.html", "ethicalads.html"]}
html_static_path = ["_static"]
html_favicon = "_static/icon.jpg"
html_logo = "_static/icon.jpg"
html_show_sourcelink = False
html_title = "Jam from Qsh.27"


# only verify ParamEMA works, low efficiency
from jamtorch.io import ParamEMA
import torch
import torch.nn as nn
import torch.nn.functional as F


class Simple(nn.Module):
    def __init__(self):
        super().__init__()
        self.a = nn.Parameter(torch.tensor([1.0, 2.0]))


def main():
    net = Simple()
    ema = ParamEMA(0.5, 8, 1)
    ema.update_parameters(net)
    for i in range(10):
        with torch.no_grad():
            net.a += 1
            ema.update_parameters(net)
        print(f"{i:04d}\t {ema.shadow_params}")

    ema_model = Simple()
    ema.copy_to(ema_model)

    print("ema\t", ema_model.a)
    print("net\t", net.a)


if __name__ == "__main__":
    main()



import itertools
from typing import List, Union

import numpy as np


def stack(arrays: List[Union[List, np.ndarray]]) -> np.ndarray:
    """stack list/ndarray of different size into ndarray with zero padding

    :param arrays: [description]
    :type arrays: List[Union[List,np.ndarray]]
    :return: stack array
    :rtype: np.ndarray
    """

    def resize(row, size):
        new = np.array(row)
        new.resize(size)
        return new

    # find longest row length
    row_length = max(arrays, key=len).__len__()
    mat = np.array([resize(row, row_length) for row in arrays])

    return mat


def stack_pad(
    arrays: List[Union[List, np.ndarray]], fillvalule: float = 0
) -> np.ndarray:
    """stack list/ndarray of different size into ndarray with zero padding

    :param arrays:
    :type arrays: List[Union[List,np.ndarray]]
    :param fillvalule: padding value, defaults to 0
    :type fillvalule: float, optional
    :return: stack array
    :rtype: np.ndarray
    """
    return np.column_stack((itertools.zip_longest(*arrays, fillvalue=fillvalule)))



from __future__ import print_function
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR
from jamtorch.trainer import Trainer
from jammy.logging import Wandb
from omegaconf import OmegaConf

cfg = OmegaConf.load("mnist.yaml")


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


# TODO: should consider process after eval
def build_loss(device):
    def loss_fn(model, feed_dict, is_train, device=device):
        data, target = feed_dict
        data, target = data.to(device), target.to(device)
        output = model(data)
        loss = F.nll_loss(output, target)
        if is_train:
            return loss, {}, {}

        pred = output.argmax(dim=1, keepdim=True)
        correct = 1.0 * pred.eq(target.view_as(pred)).sum() / len(pred)
        # fmf: off
        # from IPython import embed; embed()
        # fmf: on
        return loss, {}, {"correct": correct}

    return loss_fn


def epoch_after(trainer):
    trainer.lr_scheduler.step()
    if trainer.trainer_monitor:
        trainer.trainer_monitor.update(
            {
                "lr": trainer.optimizer.param_groups[0]["lr"],
                "epoch": trainer.epoch_cnt,
                "it": trainer.iter_cnt,
            }
        )


def main():
    # Training settings
    parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--test-batch-size",
        type=int,
        default=1000,
        metavar="N",
        help="input batch size for testing (default: 1000)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=14,
        metavar="N",
        help="number of epochs to train (default: 14)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=1.0,
        metavar="LR",
        help="learning rate (default: 1.0)",
    )
    parser.add_argument(
        "--gamma",
        type=float,
        default=0.7,
        metavar="M",
        help="Learning rate step gamma (default: 0.7)",
    )
    parser.add_argument(
        "--no-cuda", action="store_true", default=False, help="disables CUDA training"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="quickly check a single pass",
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="debug mode, no uploading"
    )
    parser.add_argument(
        "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
    )
    parser.add_argument(
        "--log-interval",
        type=int,
        default=10,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument(
        "--save-model",
        action="store_true",
        default=False,
        help="For Saving the current Model",
    )
    args = parser.parse_args()
    Wandb.launch(cfg, not args.debug)
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    train_kwargs = {"batch_size": args.batch_size}
    test_kwargs = {"batch_size": args.test_batch_size}
    if use_cuda:
        cuda_kwargs = {"num_workers": 1, "pin_memory": True, "shuffle": True}
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)

    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    dataset1 = datasets.MNIST("data", train=True, download=True, transform=transform)
    dataset2 = datasets.MNIST("data", train=False, transform=transform)
    train_loader = torch.utils.data.DataLoader(dataset1, **train_kwargs)
    test_loader = torch.utils.data.DataLoader(dataset2, **test_kwargs)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=args.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=args.gamma)

    trainer = Trainer(model, optimizer, build_loss(device), scheduler)
    trainer.set_monitor(not args.debug)
    # trainer.register_event("epoch:after", epoch_after)
    trainer.train(args.epochs, train_loader, test_loader)


if __name__ == "__main__":
    main()



"""Debug helper code for Jupyter Lab.

This module provides utilities for dumping debug information and
reproducing experiments in Jupyter notebooks.
"""

import os
from typing import Dict, Optional
import datetime

import jammy.io as jio
import jammy.utils.git as git
import ipynbname

# Constants
IMPORTANT_FILE_EXT = [".py", ".yaml", ".secret", ".yml", ".json"]


def dump_git_codebase(git_root: str, target_dir: str) -> None:
    """Dump git repository information and tracked files.

    Args:
        git_root: Path to the git repository root.
        target_dir: Directory to store the dumped information.
    """
    jio.mkdir(target_dir)

    # Create a directory for storing git repo info
    git_info_dir = os.path.join(target_dir, "git")
    jio.mkdir(git_info_dir)

    # Dump git repo info
    proj_sha, proj_diff = git.log_repo(git_root)
    with open(
        os.path.join(git_info_dir, "proj_diff.patch"), "w", encoding="utf-8"
    ) as f:
        f.write(proj_diff)
    jio.dump(proj_sha, os.path.join(git_info_dir, "sha.txt"))

    repo = git.git_repo(git_root)

    # Copy untracked files with important extensions
    for item in repo.untracked_files:
        if any(item.endswith(ext) for ext in IMPORTANT_FILE_EXT):
            jio.cp(os.path.join(git_root, item), os.path.join(target_dir, item))

    # Copy tracked files
    for item in repo.tree():
        if item.type == "blob":  # 'blob' means file in git terminology
            source_file_path = os.path.join(git_root, item.path)
            target_file_path = os.path.join(target_dir, item.path)
            os.makedirs(os.path.dirname(target_file_path), exist_ok=True)
            jio.cp(source_file_path, target_file_path)


def generate_dump_name(name: Optional[str] = None) -> str:
    """Generate a name for the dump directory.

    Args:
        name: Optional custom name for the dump.

    Returns:
        A string name for the dump directory.
    """
    if name is None:
        return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    return name


def get_target_folder(
    nb_name: str, proj_dir: str, exp_name: Optional[str] = None
) -> str:
    """Get the target folder for dumping results.

    Args:
        nb_name: Name of the Jupyter notebook.
        proj_dir: Project directory.
        exp_name: Optional experiment name.

    Returns:
        Path to the target folder.
    """
    target_root = os.path.join(proj_dir, nb_name.replace(".ipynb", ""))
    jio.mkdir(target_root)
    if exp_name is None:
        exp_name = generate_dump_name()
    return os.path.join(target_root, exp_name)


def get_notebook_path():
    """
    Get the current Jupyter notebook name and project directory.

    Returns:
        tuple: A tuple containing (notebook_name, project_directory).
               Returns (None, None) if not in a Jupyter environment.
    """
    try:
        # Get the full path of the current notebook
        notebook_path = ipynbname.path()

        if notebook_path:
            # Extract the notebook name
            notebook_name = notebook_path.name

            # Get the project directory (parent of the notebook)
            project_directory = str(notebook_path.parent)

            return notebook_name, project_directory
        else:
            print("Warning: Not running in a Jupyter notebook environment.")
            return None, None
    except Exception as e:
        print(f"Error getting notebook info: {e}")
        return None, None


def dump_jlab(
    result_dict: Optional[Dict] = None, exp_name: Optional[str] = None
) -> str:
    """Dump Jupyter Lab experiment results and codebase.

    This function is intended to be used within a Jupyter notebook.

    Args:
        result_dict: Optional dictionary of results to dump.
        exp_name: Optional experiment name.

    Returns:
        Path to the target folder where results were dumped.

    Raises:
        NotImplementedError: If unable to determine notebook name or project directory.
    """
    # TODO: Implement a way to get the current Jupyter notebook name and project directory
    # TODO: (qsh 2024-09-11) verify it!
    nb_name, proj_dir = get_notebook_path()

    if nb_name is None or proj_dir is None:
        raise NotImplementedError(
            "Unable to determine notebook name or project directory."
        )

    target_fp = get_target_folder(nb_name, proj_dir, exp_name)
    dump_git_codebase(proj_dir, os.path.join(target_fp, "codebase"))

    # Dump results
    if result_dict:
        for k, v in result_dict.items():
            jio.dump(v, os.path.join(target_fp, k))

    return target_fp



import hydra
from omegaconf import OmegaConf
import jammy.utils.hyd as hyd
import torch.multiprocessing as mp
from jamtorch import ddp as ddp

from jammy.random import reset_global_seed

def run(rank, world_size, unknown, cfg):
    # reset_global_seed(None, True)
    fn = hyd.hyd_instantiate(cfg.fn)
    fn()



@hydra.main(config_name="config.yaml")
def main(cfg):
    OmegaConf.set_struct(cfg, False)
    world_size = 2
    mp.spawn(run, args=(world_size, None, cfg), nprocs=world_size, join=True)



if __name__ == "__main__":
    main()



__all__ = ["batch_add", "batch_mul", "batch_sub", "batch_div"]


def common_broadcast(x, y):
    ndims1 = x.ndim
    ndims2 = y.ndim

    common_ndims = min(ndims1, ndims2)
    for axis in range(common_ndims):
        assert x.shape[axis] == y.shape[axis], "Dimensions not equal at axis {}".format(
            axis
        )

    if ndims1 < ndims2:
        x = x.reshape(x.shape + (1,) * (ndims2 - ndims1))
    elif ndims2 < ndims1:
        y = y.reshape(y.shape + (1,) * (ndims1 - ndims2))

    return x, y


def batch_add(x, y):
    x, y = common_broadcast(x, y)
    return x + y


def batch_mul(x, y):
    x, y = common_broadcast(x, y)
    return x * y


def batch_sub(x, y):
    x, y = common_broadcast(x, y)
    return x - y


def batch_div(x, y):
    x, y = common_broadcast(x, y)
    return x / y



import errno
import os
import os.path as osp
import re
from pathlib import Path

__all__ = ["makedirs", "glob"]


def makedirs(path):
    try:
        os.makedirs(osp.expanduser(osp.normpath(path)))
    except OSError as e:
        if e.errno != errno.EEXIST and osp.isdir(path):
            raise e


def glob(path, glob_mask="**/*", regex="", inverse=False, ftype="fd"):
    """glob files under path

    path: targeted search path
    glob_mask: simple glob pattern, use "*" when only scan current folder
    regex: re
    inverse: return not match if True
    ftype: "f" for file and "d" for dir

    Example:
        >>> glob("ckpt", regex="ckpt-[0-9]{4}\.pth", ftype="f")
        >>> ["ckpt/ckpt-0000.pth", "ckpt/ckpt-0001.pth", "ckpt/unet/ckpt-0001.pth"]
        >>> glob("ckpt", regex="ckpt-[0-9]{4}", ftype="d")
        >>> ["ckpt/ckpt-0000/", "ckpt/unet/ckpt-0000/"]
    """
    cur_path = Path(path)

    def select_type(_file):
        if "f" not in ftype:
            return osp.isdir(_file)
        if "d" not in ftype:
            return osp.isfile(_file)
        return True

    glob_files = [str(f) for f in cur_path.glob(glob_mask) if select_type(str(f))]
    if inverse:
        res = [f for f in glob_files if not re.search(regex, f)]
    else:
        res = [f for f in glob_files if re.search(regex, f)]
    return res


def mkdir_or_exist(dir_name, mode=0o777):
    if dir_name == "":
        return
    dir_name = osp.expanduser(dir_name)
    os.makedirs(dir_name, mode=mode, exist_ok=True)


def is_filepath(filepath):
    return isinstance(filepath, (str, Path))



import importlib
import os
import sys

__all__ = [
    "load_module",
    "load_module_filename",
    "load_source",
    "tuple_to_classname",
    "classname_to_tuple",
    "load_class",
    "module_vars_as_dict",
]


def load_module(module_name):
    module = importlib.import_module(module_name)
    return module


def load_module_filename(module_filename):
    assert module_filename.endswith(".py")
    realpath = os.path.realpath(module_filename)
    pos = realpath.rfind("/")
    dirname, module_name = realpath[0:pos], realpath[pos + 1 : -3]
    sys.path.insert(0, dirname)
    module = load_module(module_name)
    del sys.path[0]
    return module


def load_source(filename, name=None):
    import imp  # pylint: disable=import-self

    if name is None:
        basename = os.path.basename(filename)
        if basename.endswith(".py"):
            basename = basename[:-3]
        name = basename.replace(".", "_")

    return imp.load_source(name, filename)  # pylint: disable=arguments-out-of-order


def tuple_to_classname(t):
    assert len(t) == 2, (
        "Only tuple with length 2 (module name, class name) can be converted to classname, "
        "got {}, {}.".format(t, len(t))
    )

    return ".".join(t)


def classname_to_tuple(classname):
    pos = classname.rfind(".")
    if pos == -1:
        return "", classname
    else:
        return classname[0:pos], classname[pos + 1 :]


def load_class(classname, exit_on_error=True):
    if isinstance(classname, str):
        classname = classname_to_tuple(classname)
    elif isinstance(classname, tuple):
        assert (
            len(classname) == 2
        ), "Classname should be tuple of length 2 or a single string."
    module_name, clz_name = classname
    try:
        module = load_module(module_name)
        clz = getattr(module, clz_name)
        return clz
    except ImportError as e:
        print(
            "Cannot import {}.{}, with original error message: {}.".format(
                module_name, clz_name, str(e)
            )
        )
        if exit_on_error:
            import traceback

            traceback.print_stack()
            exit(1)
    return None


def module_vars_as_dict(module):
    res = {}
    for k in dir(module):
        if not k.startswith("__"):
            res[k] = getattr(module, k)
    return res



import copy
import ctypes

import torch


# credit
# https://github.com/pyro-ppl/pyro/blob/beddc1f1a193d7f03bd588c9d6afe389bbd2d10c/pyro/distributions/util.py
class _DeepToMemo(dict):
    def __init__(self, to_args, to_kwargs):
        super().__init__()
        self._to_args = to_args
        self._to_kwargs = to_kwargs

    def get(self, key, default=None):
        result = super().get(key, default)
        if result is default:
            # Assume key is the id of another object, and look up that object.
            old = ctypes.cast(key, ctypes.py_object).value
            if isinstance(
                old, (torch.Tensor, torch.nn.Module)
            ):  # or maybe duck type here?
                self[key] = result = old.to(*self._to_args, **self._to_kwargs)
        return result


def deep_to(obj, *args, **kwargs):
    r"""
    Create a deep copy of an arbitrary Python object, calling ``.to(*args,
    **kwargs)`` on all :class:`torch.Tensor` s and :class:`torch.nn.Module` s
    in that object.

    Like :meth:`torch.Tensor.to` but unlike :meth:`torch.nn.Module.to`, this
    creates new objects. For compatibility with existing PyTorch module
    behavior, this first calls :meth:`torch.nn.Module.to` in-place, then
    creates a deep copy of the module.

    :param obj: Any python object.
    :param \*args:
    :param \*\*kwargs: See :meth:`torch.Tensor.to`.
    :returns: A deep copy of ``obj`` with all :class:`torch.Tensor` s and
        :class:`torch.nn.Module` s mapped.
    """
    memo = _DeepToMemo(args, kwargs)
    return copy.deepcopy(obj, memo)



# pylint: disable=redefined-outer-name
"""Tests for TensorSparseSample class.

This module contains comprehensive tests for the TensorSparseSample class,
covering initialization, updates, and various sampling strategies.
"""

import pytest
import torch
from torch import Tensor

from .distribution_sample import TensorSparseSample, get_tensor_stat_info


def test_get_stat_info():
    """Tests the get_stat_info function."""
    tensor = torch.randn(1000)
    stat_info = get_tensor_stat_info(tensor, "Random")
    print(stat_info)


# Test fixtures
@pytest.fixture
def small_1d_tensor() -> Tensor:
    """Creates a small 1D tensor for basic tests."""
    return torch.arange(10, dtype=torch.float32)


@pytest.fixture
def large_1d_tensor() -> Tensor:
    """Creates a larger 1D tensor for testing sampling strategies."""
    return torch.arange(1000, dtype=torch.float32)


@pytest.fixture
def tensor_2d() -> Tensor:
    """Creates a 2D tensor for testing multi-dimensional cases."""
    return torch.arange(100, dtype=torch.float32).reshape(10, 10)


@pytest.fixture
def tensor_3d() -> Tensor:
    """Creates a 3D tensor for testing multi-dimensional cases."""
    return torch.arange(120, dtype=torch.float32).reshape(4, 5, 6)


@pytest.fixture
def basic_sparse_sample(small_1d_tensor: Tensor) -> TensorSparseSample:
    """Creates a basic TensorSparseSample instance."""
    return TensorSparseSample.from_tensor(small_1d_tensor, sample_size=5)


class TestTensorSparseSampleInitialization:
    """Tests for TensorSparseSample initialization."""

    def test_basic_initialization(self):
        """Tests basic initialization with valid inputs."""
        indices = torch.tensor([0, 1, 2])
        sample = TensorSparseSample(indices=indices)
        assert torch.equal(sample.indices, indices)
        assert sample.values is None
        assert sample.ntensor == 0

    def test_initialization_with_values(self):
        """Tests initialization with both indices and values."""
        indices = torch.tensor([0, 1, 2])
        values = torch.tensor([[1.0, 2.0, 3.0]])  # Shape: (1, 3)
        sample = TensorSparseSample(indices=indices, values=values, ntensor=1)
        assert torch.equal(sample.indices, indices)
        assert torch.equal(sample.values, values)
        assert sample.ntensor == 1

    def test_wrong_number_of_values(self):
        """Tests initialization with wrong number of values."""
        indices = torch.tensor([0, 1, 2])
        values = torch.tensor([[1.0, 2.0]])  # Wrong number of values
        with pytest.raises(ValueError, match="Number of values .* must match"):
            TensorSparseSample(indices=indices, values=values)

    def test_wrong_shape_values(self):
        """Tests initialization with wrong shape values (1D instead of 2D)."""
        indices = torch.tensor([0, 1, 2])
        values = torch.tensor([1.0, 2.0, 3.0])  # 1D instead of 2D
        with pytest.raises(ValueError, match="values must be a 2D tensor"):
            TensorSparseSample(indices=indices, values=values)

    def test_wrong_type_values(self):
        """Tests initialization with wrong type values."""
        indices = torch.tensor([0, 1, 2])
        values = [1.0, 2.0, 3.0]  # List instead of tensor
        with pytest.raises(TypeError, match="values must be a torch.Tensor"):
            TensorSparseSample(indices=indices, values=values)

    def test_wrong_type_indices(self):
        """Tests initialization with wrong type indices."""
        indices = [0, 1, 2]  # List instead of tensor
        with pytest.raises(TypeError, match="indices must be a torch.Tensor"):
            TensorSparseSample(indices=indices)


class TestTensorSparseSampleUpdate:
    """Tests for update functionality."""

    def test_single_update(self, small_1d_tensor: Tensor):
        """Tests single update behavior."""
        sample = TensorSparseSample.from_tensor(small_1d_tensor, sample_size=3)
        assert sample.ntensor == 1
        assert sample.values.shape == (1, 3)
        assert torch.equal(sample.values[0], small_1d_tensor[sample.indices])

    def test_multiple_updates(self, small_1d_tensor: Tensor):
        """Tests multiple update behavior."""
        sample = TensorSparseSample.from_tensor(small_1d_tensor, sample_size=3)

        # Second update
        new_tensor = small_1d_tensor + 10
        sample.update(new_tensor)
        assert sample.ntensor == 2
        assert sample.values.shape == (2, 3)
        assert torch.equal(sample.values[1], new_tensor[sample.indices])

        # Third update
        newer_tensor = small_1d_tensor + 20
        sample.update(newer_tensor)
        assert sample.ntensor == 3
        assert sample.values.shape == (3, 3)
        assert torch.equal(sample.values[2], newer_tensor[sample.indices])

    def test_update_different_dim_tensor(self, basic_sparse_sample: TensorSparseSample):
        """Tests update with tensor of different dimensions."""
        tensor_2d = torch.randn(10, 10)
        with pytest.raises(ValueError):
            basic_sparse_sample.update(tensor_2d)

    def test_update_value_consistency(self, small_1d_tensor: Tensor):
        """Tests that values are consistent across updates."""
        sample = TensorSparseSample.from_tensor(small_1d_tensor, sample_size=3)
        values_history = [sample.values[0].clone()]

        for i in range(3):
            new_tensor = small_1d_tensor + (i + 1) * 10
            sample.update(new_tensor)
            values_history.append(sample.values[i + 1].clone())

        # Check each update matches expected values
        for i, values in enumerate(values_history):
            expected_tensor = small_1d_tensor + i * 10
            assert torch.allclose(values, expected_tensor[sample.indices])


class TestSamplingStrategies:
    """Tests for different sampling strategies."""

    @pytest.mark.parametrize("strategy", ["random", "uniform", "edges"])
    def test_sampling_strategies(self, large_1d_tensor: Tensor, strategy: str):
        """Tests different sampling strategies."""
        sample_size = 10
        sample = TensorSparseSample.from_tensor(
            large_1d_tensor, sample_size=sample_size, strategy=strategy
        )

        assert len(sample) == sample_size
        assert sample.values.shape == (1, sample_size)
        assert torch.all(sample.indices < len(large_1d_tensor))

    def test_edges_strategy_distribution(self, large_1d_tensor: Tensor):
        """Tests that edges strategy samples from start, middle, and end."""
        sample = TensorSparseSample.from_tensor(
            large_1d_tensor, sample_size=9, strategy="edges"
        )

        indices = sample.indices.cpu().numpy()
        n = len(large_1d_tensor)

        # Check if we have samples from each third
        assert any(idx < n / 3 for idx in indices)  # Start
        assert any(n / 3 <= idx < 2 * n / 3 for idx in indices)  # Middle
        assert any(idx >= 2 * n / 3 for idx in indices)  # End

    def test_uniform_strategy_spacing(self, large_1d_tensor: Tensor):
        """Tests that uniform strategy has consistent spacing."""
        sample = TensorSparseSample.from_tensor(
            large_1d_tensor, sample_size=5, strategy="uniform"
        )

        diffs = sample.indices[1:] - sample.indices[:-1]
        assert torch.allclose(diffs, diffs[0] * torch.ones_like(diffs))


class TestMultiDimensionalTensors:
    """Tests for handling multi-dimensional tensors."""

    @pytest.mark.parametrize("tensor_fixture", ["tensor_2d", "tensor_3d"])
    def test_multi_dim_sampling(self, tensor_fixture: str, request):
        """Tests sampling from multi-dimensional tensors."""
        tensor = request.getfixturevalue(tensor_fixture)
        sample = TensorSparseSample.from_tensor(tensor, sample_size=5)

        # Check indices shape
        assert sample.indices.shape[0] == 5  # number of samples
        assert sample.indices.shape[1] == tensor.dim()  # number of dimensions
        assert sample.values.shape == (1, 5)  # (ntensor, n_samples)

        # Verify values match
        values = []
        for idx in sample.indices:
            values.append(tensor[tuple(idx.tolist())])
        expected_values = torch.stack(values)
        assert torch.equal(sample.values[0], expected_values)

    def test_multi_dim_updates(self, tensor_2d: Tensor):
        """Tests updates with multi-dimensional tensors."""
        sample = TensorSparseSample.from_tensor(tensor_2d, sample_size=5)
        new_tensor = tensor_2d + 10

        sample.update(new_tensor)
        assert sample.ntensor == 2
        assert sample.values.shape == (2, 5)

        # Verify new values match
        values = []
        for idx in sample.indices:
            values.append(new_tensor[tuple(idx.tolist())])
        expected_values = torch.stack(values)
        assert torch.equal(sample.values[1], expected_values)


class TestErrorHandling:
    """Tests for error handling."""

    def test_invalid_strategy(self, small_1d_tensor: Tensor):
        """Tests error handling for invalid sampling strategy."""
        with pytest.raises(ValueError, match="Unknown sampling strategy"):
            TensorSparseSample.from_tensor(small_1d_tensor, strategy="invalid_strategy")

    def test_sample_size_too_large(self, small_1d_tensor: Tensor):
        """Tests error handling for too large sample size."""
        with pytest.raises(ValueError, match="sample_size .* cannot be larger"):
            TensorSparseSample.from_tensor(
                small_1d_tensor, sample_size=len(small_1d_tensor) + 1
            )

    def test_out_of_bounds_indices(self, small_1d_tensor: Tensor):
        """Tests error handling for out-of-bounds indices."""
        invalid_indices = torch.tensor([len(small_1d_tensor) + 1])
        with pytest.raises(ValueError, match="Index out of bounds"):
            TensorSparseSample.from_tensor(small_1d_tensor, indices=invalid_indices)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) OpenMMLab. All rights reserved.
import pickle
from io import BytesIO
from typing import Any

from .base import BaseFileHandler


class PickleHandler(BaseFileHandler):

    str_like = False

    def load_from_fileobj(self, file: BytesIO, **kwargs):
        return pickle.load(file, **kwargs)

    def load_from_path(self, filepath, **kwargs):
        return super().load_from_path(filepath, mode="rb", **kwargs)

    def dump_to_str(self, obj, **kwargs):
        kwargs.setdefault("protocol", 2)
        return pickle.dumps(obj, **kwargs)

    def dump_to_fileobj(self, obj: Any, file: BytesIO, **kwargs):
        kwargs.setdefault("protocol", 2)
        pickle.dump(obj, file, **kwargs)

    def dump_to_path(self, obj, filepath, **kwargs):
        # super().dump_to_path(obj, filepath, mode="wb", **kwargs)
        with open(filepath, "wb") as f:
            pickle.dump(obj, f, **kwargs)



from jammy.utils.meta import Singleton
import threading
import time
import random
import os
import multiprocessing as mp


class Earth(metaclass=Singleton):
    def __init__(self, name):
        self.name = name

    def __call__(self):
        print(threading.get_ident(), self.name)


def test_global():
    this_earth = Earth("abc")
    that_earth = Earth("def")
    this_earth()
    that_earth()


def create(name):
    earth = Earth(name)
    t_time = random.randint(1, 5)
    print(os.getpid(), threading.get_ident(), "sleep: ", t_time)
    time.sleep(t_time)
    earth()


def test_thread():
    N = 10
    thread_l = []
    for i in range(N):
        thread_l.append(threading.Thread(target=create, args=[f"THD Naming {i}"]))
    for i in range(N):
        thread_l[i].start()
    for i in range(N):
        thread_l[i].join()


def test_mp():
    N = 10
    mp_l = []
    for i in range(N):
        mp_l.append(mp.Process(target=create, args=[f"MP Naming {i}"]))
    for i in range(N):
        mp_l[i].start()
    for i in range(N):
        mp_l[i].join()


if __name__ == "__main__":
    test_global()
    Singleton._instances.clear()
    print("*" * 100)
    test_thread()
    Singleton._instances.clear()
    print("*" * 100)
    test_mp()



# pylint: skip-file
import time
from multiprocessing import Pool

from jammy.utils.filelock import try_filelock, TryFileLock

w_file = "writing.log"


def work(pid):
    global w_file

    def fn():
        with open(w_file, "a") as f:
            f.write(f"{pid}\n")
        print(f"{pid} start sleep")
        time.sleep(1)
        a = 1.0
        b = 1
        c = a / b
        print(f"{pid} end sleep")

    try_filelock(
        fn,
        w_file,
        is_soft=False,
        timeout=10,
    )

def work2(pid):
    global w_file
    with TryFileLock(w_file, is_soft=False, timeout=1) as flock:
        print(pid, flock)
        if flock is None:
            return
        with open(w_file, "a") as f:
            f.write(f"{pid}\n")
        print(f"{pid} start sleep")
        time.sleep(10)
        a = 1.0
        b = 1
        print(f"{pid} end sleep")

if __name__ == "__main__":
    # with Pool(5) as p:
    #     print(p.map(work, [1, 2, 3, 4, 5]))
    print("===" * 10)
    with Pool(5) as p:
        print(p.map(work2, [1, 2, 3, 4, 5]))



"""Implementations of discrete distributions."""

import torch
from torch.nn import functional as F

from .base import Distribution
import jamtorch.utils.ts_ops as ts_ops


class ConditionalIndependentBernoulli(Distribution):
    """An independent Bernoulli whose parameters are functions of a context."""

    def __init__(self, shape, context_encoder=None):
        """Constructor.

        Args:
            shape: list, tuple or torch.Size, the shape of the input variables.
            context_encoder: callable or None, encodes the context to the distribution parameters.
                If None, defaults to the identity function.
        """
        super().__init__()
        self._shape = torch.Size(shape)
        if context_encoder is None:
            self._context_encoder = lambda x: x
        else:
            self._context_encoder = context_encoder

    def _compute_params(self, context):
        """Compute the logits from context."""
        if context is None:
            raise ValueError("Context can't be None.")

        logits = self._context_encoder(context)
        if logits.shape[0] != context.shape[0]:
            raise RuntimeError(
                "The batch dimension of the parameters is inconsistent with the input."
            )

        return logits.reshape(logits.shape[0], *self._shape)

    def _log_prob(self, inputs, context):
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )

        # Compute parameters.
        logits = self._compute_params(context)
        assert logits.shape == inputs.shape

        # Compute log prob.
        log_prob = -inputs * F.softplus(-logits) - (1.0 - inputs) * F.softplus(logits)
        log_prob = ts_ops.sum_except_batch(log_prob, num_batch_dims=1)
        return log_prob

    def _sample(self, num_samples, context):
        # Compute parameters.
        logits = self._compute_params(context)
        probs = torch.sigmoid(logits)
        probs = ts_ops.repeat_rows(probs, num_samples)

        # Generate samples.
        context_size = context.shape[0]
        noise = torch.rand(context_size * num_samples, *self._shape)
        samples = (noise < probs).float()
        return ts_ops.split_leading_dim(samples, [context_size, num_samples])

    def _mean(self, context):
        logits = self._compute_params(context)
        return torch.sigmoid(logits)



import atexit
import os
import sys

from loguru._logger import Core, Logger


def add_relative_path(record):
    start = os.getcwd()
    record["extra"]["relative_path"] = os.path.relpath(record["file"].path, start)


logger = Logger(
    core=Core(),
    exception=None,
    depth=1,
    record=False,
    lazy=False,
    colors=False,
    raw=False,
    capture=True,
    patchers=[add_relative_path],
    extra={},
)

atexit.register(logger.remove)

__all__ = [
    "get_logger",
    "LOG_FORMAT",
]

logger.remove()

LOGGER_SINK = {}
LOG_FORMAT = (
    "[<green>{time:MM-DD HH:mm:ss}</green>|"
    "<red>{process.name}</red>|"
    "<level>{level: <8}</level>|"
    "<cyan>{file.path}</cyan>:<cyan>{line}</cyan>:<cyan>{function}</cyan>]"
    "<level>{message}</level>"
)


def get_logger(file_name=None, clear=False, **kwargs):
    global LOGGER_SINK  # pylint: disable=global-statement,global-variable-not-assigned
    if clear:
        logger.remove()
        LOGGER_SINK.clear()
    if file_name is None:
        file_name = sys.stderr
        if "level" not in kwargs:
            kwargs["level"] = "INFO"
    if "format" not in kwargs:
        kwargs["format"] = LOG_FORMAT
    if file_name in LOGGER_SINK.values():
        # FIXME:
        # logger.debug("already registered")
        logger.debug(f"{str(file_name)} already registered")
    else:
        # if "level" not in kwargs:
        # kwargs["level"] = "DEBUG" if jam_is_debug() else "INFO"
        sink_id = logger.add(file_name, **kwargs)
        LOGGER_SINK[sink_id] = file_name
        logger.debug(f"sink_id: {sink_id:02d} ---> {str(file_name)}")
    return logger



import functools

import torch

__all__ = [
    "no_grad_func",
    "disable_grad",
]


def no_grad_func(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        with torch.no_grad():
            return func(*args, **kwargs)

    return new_func


def disable_grad(module):
    for param in module.parameters():
        param.requires_grad = False



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : mem.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

from .kv import KVStoreBase


class MemKVStore(KVStoreBase):
    def __init__(self, readonly=False):
        super().__init__(readonly=readonly)
        self._store = dict()

    def _has(self, key):
        return key in self._store

    def _get(self, key, default):
        return self._store.get(key, default)

    def _put(self, key, value, replace):
        if not replace:
            self._store.setdefault(key, value)
        else:
            self._store[key] = value

    def _erase(self, key):
        return self._store.pop(key)

    def _keys(self):
        return self._store.keys()



# pylint: disable=import-outside-toplevel

from typing import IO, Optional, Tuple, Union

import numpy as np

from .base import BaseFileHandler

try:
    from PIL import Image
except ImportError:
    Image = None


class PILHandler(BaseFileHandler):
    str_like = False

    def load_from_fileobj(
        self,
        file: IO[bytes],
        fmt: str = "pil",
        size: Optional[Union[int, Tuple[int, int]]] = None,
        **kwargs,
    ):
        """
        Load an image from a file-like object and return it in a specified format.

        Args:
            file (IO[bytes]): A file-like object containing the image data.
            fmt (str): The format to convert the image into. Options are \
                'numpy', 'np', 'npy', 'type' (all return numpy arrays), \
                    'pil' (returns PIL Image), 'th', 'torch' (returns a torch tensor).
            size (Optional[Union[int, Tuple[int, int]]]): The new size of the image as a single integer \
                or a tuple of (width, height). If specified, the image is resized accordingly.
            **kwargs: Additional keyword arguments that can be passed to conversion functions.

        Returns:
            Image data in the format specified by `fmt`.

        Raises:
            IOError: If the image cannot be loaded or processed.
            ValueError: If the specified format is unsupported.
        """
        try:
            img = Image.open(file)
            img.load()  # Explicitly load the image data
            if size is not None:
                if isinstance(size, int):
                    size = (
                        size,
                        size,
                    )  # create a tuple if only one integer is provided
                img = img.resize(size, Image.ANTIALIAS)

            # Return the image in the requested format
            if fmt in ["numpy", "np", "npy"]:
                return np.array(img, **kwargs)
            if fmt == "pil":
                return img
            if fmt in ["th", "torch"]:
                import torch

                # Convert to tensor
                img_tensor = torch.from_numpy(np.array(img, **kwargs))
                # Convert image from HxWxC to CxHxW
                if img_tensor.ndim == 3:
                    img_tensor = img_tensor.permute(2, 0, 1)
                return img_tensor
            raise ValueError(
                "Unsupported format. Supported formats are 'numpy', 'np', 'npy', 'pil', 'th', and 'torch'."
            )
        except Exception as e:
            raise IOError(f"Unable to load image: {e}") from e

    def dump_to_fileobj(self, obj, file: IO[bytes], **kwargs):
        obj.save(file, **kwargs)

    def dump_to_str(self, obj, **kwargs):
        raise NotImplementedError



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : cache.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

import collections
import functools
import os.path as osp
import threading
from typing import Callable

from jammy.logging import get_logger

from .meta import synchronized

logger = get_logger()

__all__ = ["cached_property", "cached_result", "fs_cached_result"]


def cached_property(fget):
    """A decorator that converts a function into a cached property. Similar to ``@property``, but the function result is cached. This function has threading lock support."""

    mutex = collections.defaultdict(threading.Lock)
    cache = dict()

    def impl(self):
        nonlocal impl
        id_ = id(self)
        with mutex[id_]:
            if id_ not in cache:
                cache[id_] = fget(self)
                return cache[id_]
            else:
                return cache[id_]

    return property(impl)


def cached_result(func):
    """A decorator that caches the result of a function. Note that this decorator does not support any arguments to the function."""

    def impl():
        nonlocal impl
        ret = func()
        impl = lambda: ret
        return ret

    @synchronized()
    @functools.wraps(func)
    def f():
        return impl()

    return f


def fs_cached_result(
    filename: str, force_update: bool = False, verbose: bool = False
) -> Callable[[Callable], Callable]:
    """A decorator that caches the result of a function into a file. Note that this decorator does not take care of any arguments to the function.

    Args:
        filename: the filename to store the result.
        force_update: if True, the result will be updated even if the file exists.
        verbose: if True, the filenames will be printed to the console.

    Returns:
        a decorator function.
    """
    import jammy.io as io

    def wrapper(func):
        @synchronized()
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if not force_update and osp.exists(filename):
                if verbose:
                    logger.info('Using cached results from "{}".'.format(filename))
                cached_value = io.load(filename)
                if cached_value is not None:
                    return cached_value

            computed_value = func(*args, **kwargs)
            if verbose:
                logger.info('Writing result cache to "{}".'.format(filename))
            io.dump(filename, computed_value)
            return computed_value

        return wrapped_func

    return wrapper



# pylint: skip-file
import jammy
from jammy import *
from jammy import io as jio

parser = jammy.cli.JamArgumentParser()
parser.add_argument("filename", nargs="+")
args = parser.parse_args()


def simple():
    for i, filename in enumerate(args.filename):
        globals()[f"f{i + 1}"] = jio.load(filename)

    from IPython import embed

    embed()


def simple_print():
    for i, filename in enumerate(args.filename):
        print(f"############## {i:03d} \t {filename} ##############")
        stprint(jio.load(filename))



# Copyright (c) OpenMMLab. All rights reserved.
import io
import os
import os.path as osp
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Iterator, Optional, Tuple, Union

from jammy.io.path import mkdir_or_exist

from .base import BaseStorageBackend


class LocalBackend(BaseStorageBackend):
    """Raw local storage backend."""

    _allow_symlink = True

    def get(self, filepath: Union[str, Path]) -> bytes:
        """Read bytes from a given ``filepath`` with 'rb' mode.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes: Expected bytes object.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.get(filepath)
            b'hello world'
        """
        with open(filepath, "rb") as f:
            value = f.read()
        return value

    def get_text(self, filepath: Union[str, Path], encoding: str = "utf-8") -> str:
        """Read text from a given ``filepath`` with 'r' mode.

        Args:
            filepath (str or Path): Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Defaults to 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.get_text(filepath)
            'hello world'
        """
        with open(filepath, encoding=encoding) as f:
            text = f.read()
        return text

    def put(self, obj: Union[bytes, io.BytesIO], filepath: Union[str, Path]) -> None:
        """Write bytes to a given ``filepath`` with 'wb' mode.

        Note:
            ``put`` will create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (bytes): Data to be written.
            filepath (str or Path): Path to write data.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.put(b'hello world', filepath)
        """
        mkdir_or_exist(osp.dirname(filepath))
        if isinstance(obj, io.BytesIO):
            obj.seek(0)
            obj = obj.getvalue()
        with open(filepath, "wb") as f:
            f.write(obj)

    def put_text(
        self, obj: str, filepath: Union[str, Path], encoding: str = "utf-8"
    ) -> None:
        """Write text to a given ``filepath`` with 'w' mode.

        Note:
            ``put_text`` will create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str): The encoding format used to open the ``filepath``.
                Defaults to 'utf-8'.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.put_text('hello world', filepath)
        """
        mkdir_or_exist(osp.dirname(filepath))
        with open(filepath, "w", encoding=encoding) as f:
            f.write(obj)

    def exists(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.exists(filepath)
            True
        """
        return osp.exists(filepath)

    def isdir(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a directory.

        Args:
            filepath (str or Path): Path to be checked whether it is a
                directory.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a directory,
            ``False`` otherwise.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/dir'
            >>> backend.isdir(filepath)
            True
        """
        return osp.isdir(filepath)

    def isfile(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a file.

        Args:
            filepath (str or Path): Path to be checked whether it is a file.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a file, ``False``
            otherwise.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.isfile(filepath)
            True
        """
        return osp.isfile(filepath)

    def join_path(
        self, filepath: Union[str, Path], *filepaths: Union[str, Path]
    ) -> str:
        r"""Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of \*filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath1 = '/path/of/dir1'
            >>> filepath2 = 'dir2'
            >>> filepath3 = 'path/of/file'
            >>> backend.join_path(filepath1, filepath2, filepath3)
            '/path/of/dir/dir2/path/of/file'
        """
        # TODO, if filepath or filepaths are Path, should return Path
        return osp.join(filepath, *filepaths)

    @contextmanager
    def get_local_path(
        self,
        filepath: Union[str, Path],
    ) -> Generator[Union[str, Path], None, None]:
        """Only for unified API and do nothing.

        Args:
            filepath (str or Path): Path to be read data.
            backend_args (dict, optional): Arguments to instantiate the
                corresponding backend. Defaults to None.

        Examples:
            >>> backend = LocalBackend()
            >>> with backend.get_local_path('s3://bucket/abc.jpg') as path:
            ...     # do something here
        """
        yield filepath

    def copyfile(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Copy a file src to dst and return the destination file.

        src and dst should have the same prefix. If dst specifies a directory,
        the file will be copied into dst using the base filename from src. If
        dst specifies a file that already exists, it will be replaced.

        Args:
            src (str or Path): A file to be copied.
            dst (str or Path): Copy file to dst.

        Returns:
            str: The destination file.

        Raises:
            SameFileError: If src and dst are the same file, a SameFileError
                will be raised.

        Examples:
            >>> backend = LocalBackend()
            >>> # dst is a file
            >>> src = '/path/of/file'
            >>> dst = '/path1/of/file1'
            >>> # src will be copied to '/path1/of/file1'
            >>> backend.copyfile(src, dst)
            '/path1/of/file1'

            >>> # dst is a directory
            >>> dst = '/path1/of/dir'
            >>> # src will be copied to '/path1/of/dir/file'
            >>> backend.copyfile(src, dst)
            '/path1/of/dir/file'
        """
        return shutil.copy(src, dst)

    def copytree(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Recursively copy an entire directory tree rooted at src to a
        directory named dst and return the destination directory.

        src and dst should have the same prefix and dst must not already exist.

        TODO: Whether to support dirs_exist_ok parameter.

        Args:
            src (str or Path): A directory to be copied.
            dst (str or Path): Copy directory to dst.

        Returns:
            str: The destination directory.

        Raises:
            FileExistsError: If dst had already existed, a FileExistsError will
                be raised.

        Examples:
            >>> backend = LocalBackend()
            >>> src = '/path/of/dir1'
            >>> dst = '/path/of/dir2'
            >>> backend.copytree(src, dst)
            '/path/of/dir2'
        """
        return shutil.copytree(src, dst)

    def copyfile_from_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Copy a local file src to dst and return the destination file. Same
        as :meth:`copyfile`.

        Args:
            src (str or Path): A local file to be copied.
            dst (str or Path): Copy file to dst.

        Returns:
            str: If dst specifies a directory, the file will be copied into dst
            using the base filename from src.

        Raises:
            SameFileError: If src and dst are the same file, a SameFileError
                will be raised.

        Examples:
            >>> backend = LocalBackend()
            >>> # dst is a file
            >>> src = '/path/of/file'
            >>> dst = '/path1/of/file1'
            >>> # src will be copied to '/path1/of/file1'
            >>> backend.copyfile_from_local(src, dst)
            '/path1/of/file1'

            >>> # dst is a directory
            >>> dst = '/path1/of/dir'
            >>> # src will be copied to
            >>> backend.copyfile_from_local(src, dst)
            '/path1/of/dir/file'
        """
        return self.copyfile(src, dst)

    def copytree_from_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Recursively copy an entire directory tree rooted at src to a
        directory named dst and return the destination directory. Same as
        :meth:`copytree`.

        Args:
            src (str or Path): A local directory to be copied.
            dst (str or Path): Copy directory to dst.

        Returns:
            str: The destination directory.

        Examples:
            >>> backend = LocalBackend()
            >>> src = '/path/of/dir1'
            >>> dst = '/path/of/dir2'
            >>> backend.copytree_from_local(src, dst)
            '/path/of/dir2'
        """
        return self.copytree(src, dst)

    def copyfile_to_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Copy the file src to local dst and return the destination file. Same
        as :meth:`copyfile`.

        If dst specifies a directory, the file will be copied into dst using
        the base filename from src. If dst specifies a file that already
        exists, it will be replaced.

        Args:
            src (str or Path): A file to be copied.
            dst (str or Path): Copy file to to local dst.

        Returns:
            str: If dst specifies a directory, the file will be copied into dst
            using the base filename from src.

        Examples:
            >>> backend = LocalBackend()
            >>> # dst is a file
            >>> src = '/path/of/file'
            >>> dst = '/path1/of/file1'
            >>> # src will be copied to '/path1/of/file1'
            >>> backend.copyfile_to_local(src, dst)
            '/path1/of/file1'

            >>> # dst is a directory
            >>> dst = '/path1/of/dir'
            >>> # src will be copied to
            >>> backend.copyfile_to_local(src, dst)
            '/path1/of/dir/file'
        """
        return self.copyfile(src, dst)

    def copytree_to_local(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> str:
        """Recursively copy an entire directory tree rooted at src to a local
        directory named dst and return the destination directory.

        Args:
            src (str or Path): A directory to be copied.
            dst (str or Path): Copy directory to local dst.
            backend_args (dict, optional): Arguments to instantiate the
                prefix of uri corresponding backend. Defaults to None.

        Returns:
            str: The destination directory.

        Examples:
            >>> backend = LocalBackend()
            >>> src = '/path/of/dir1'
            >>> dst = '/path/of/dir2'
            >>> backend.copytree_from_local(src, dst)
            '/path/of/dir2'
        """
        return self.copytree(src, dst)

    def remove(self, filepath: Union[str, Path]) -> None:
        """Remove a file.

        Args:
            filepath (str or Path): Path to be removed.

        Raises:
            IsADirectoryError: If filepath is a directory, an IsADirectoryError
                will be raised.
            FileNotFoundError: If filepath does not exist, an FileNotFoundError
                will be raised.

        Examples:
            >>> backend = LocalBackend()
            >>> filepath = '/path/of/file'
            >>> backend.remove(filepath)
        """
        if not self.exists(filepath):
            raise FileNotFoundError(f"filepath {filepath} does not exist")

        if self.isdir(filepath):
            raise IsADirectoryError("filepath should be a file")

        os.remove(filepath)

    def rmtree(self, dir_path: Union[str, Path]) -> None:
        """Recursively delete a directory tree.

        Args:
            dir_path (str or Path): A directory to be removed.

        Examples:
            >>> dir_path = '/path/of/dir'
            >>> backend.rmtree(dir_path)
        """
        shutil.rmtree(dir_path)

    def copy_if_symlink_fails(
        self,
        src: Union[str, Path],
        dst: Union[str, Path],
    ) -> bool:
        """Create a symbolic link pointing to src named dst.

        If failed to create a symbolic link pointing to src, directly copy src
        to dst instead.

        Args:
            src (str or Path): Create a symbolic link pointing to src.
            dst (str or Path): Create a symbolic link named dst.

        Returns:
            bool: Return True if successfully create a symbolic link pointing
            to src. Otherwise, return False.

        Examples:
            >>> backend = LocalBackend()
            >>> src = '/path/of/file'
            >>> dst = '/path1/of/file1'
            >>> backend.copy_if_symlink_fails(src, dst)
            True
            >>> src = '/path/of/dir'
            >>> dst = '/path1/of/dir1'
            >>> backend.copy_if_symlink_fails(src, dst)
            True
        """
        try:
            os.symlink(src, dst)
            return True
        except Exception:
            if self.isfile(src):
                self.copyfile(src, dst)
            else:
                self.copytree(src, dst)
            return False

    def list_dir_or_file(
        self,
        dir_path: Union[str, Path],
        list_dir: bool = True,
        list_file: bool = True,
        suffix: Optional[Union[str, Tuple[str]]] = None,
        recursive: bool = False,
    ) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

        Args:
            dir_path (str or Path): Path of the directory.
            list_dir (bool): List the directories. Defaults to True.
            list_file (bool): List the path of files. Defaults to True.
            suffix (str or tuple[str], optional): File suffix that we are
                interested in. Defaults to None.
            recursive (bool): If set to True, recursively scan the directory.
                Defaults to False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.

        Examples:
            >>> backend = LocalBackend()
            >>> dir_path = '/path/of/dir'
            >>> # list those files and directories in current directory
            >>> for file_path in backend.list_dir_or_file(dir_path):
            ...     print(file_path)
            >>> # only list files
            >>> for file_path in backend.list_dir_or_file(dir_path, list_dir=False):
            ...     print(file_path)
            >>> # only list directories
            >>> for file_path in backend.list_dir_or_file(dir_path, list_file=False):
            ...     print(file_path)
            >>> # only list files ending with specified suffixes
            >>> for file_path in backend.list_dir_or_file(dir_path, suffix='.txt'):
            ...     print(file_path)
            >>> # list all files and directory recursively
            >>> for file_path in backend.list_dir_or_file(dir_path, recursive=True):
            ...     print(file_path)
        """  # noqa: E501
        if list_dir and suffix is not None:
            raise TypeError("`suffix` should be None when `list_dir` is True")

        if (suffix is not None) and not isinstance(suffix, (str, tuple)):
            raise TypeError("`suffix` must be a string or tuple of strings")

        root = dir_path

        def _list_dir_or_file(dir_path, list_dir, list_file, suffix, recursive):
            for entry in os.scandir(dir_path):
                if not entry.name.startswith(".") and entry.is_file():
                    rel_path = osp.relpath(entry.path, root)
                    if (suffix is None or rel_path.endswith(suffix)) and list_file:
                        yield rel_path
                elif osp.isdir(entry.path):
                    if list_dir:
                        rel_dir = osp.relpath(entry.path, root)
                        yield rel_dir
                    if recursive:
                        yield from _list_dir_or_file(
                            entry.path, list_dir, list_file, suffix, recursive
                        )

        return _list_dir_or_file(dir_path, list_dir, list_file, suffix, recursive)



from jammy.logging import Wandb
import torch.nn.functional as F
import hydra
import torch
from omegaconf import OmegaConf
import jamtorch.ddp.ddp_utils as ddp_utils
from jamtorch.ddp.ddp_trainer import DDPTrainer
from jamtorch.trainer.amp import fp16_wrapper
import jammy.utils.hyd as hyd
import torch.multiprocessing as mp
# from torch.multiprocessing import Process

@fp16_wrapper
class Trainer(DDPTrainer):
    pass

def loss_wrapper(device):
    def loss_fn(model, feed_dict, is_train):
        data, label = feed_dict
        data, label = data.to(device), label.to(device)
        pred = model(data)
        loss = F.cross_entropy(pred, label)
        return loss, {}, {}
    return loss_fn


@ddp_utils.ddp_runner
def run(cfg):
    train_set, val_set = hyd.hyd_instantiate(cfg.data.datasets)()
    train_loader, train_sampler, val_loader, val_sampler =\
        hyd.hyd_instantiate(cfg.data.dataloader, train_set, val_set, \
            rank=cfg.trainer.rank, world_size=cfg.trainer.world_size)()
    model = hyd.hyd_instantiate(cfg.model)
    device = torch.device(f"cuda:{cfg.trainer.gpu}")
    trainer = Trainer(cfg.trainer, loss_wrapper(device))
    optimizer = hyd.hyd_instantiate(cfg.optimizer, model.parameters())
    trainer.set_model_optim(model, optimizer)
    trainer.set_dataloader(train_loader, val_loader)
    trainer.set_sampler(train_sampler, val_sampler)
    trainer.train()

@ddp_utils.ddp_runner
def debug_random(cfg):
    print(cfg.trainer.gpu, torch.randn(size=(3,)))
    torch.distributed.barrier()


@hydra.main(config_path="conf/", config_name="config.yaml")
def main(cfg):
    # Wandb.launch(cfg, cfg.trainer.wandb.log, True)
    OmegaConf.set_struct(cfg, False)
    world_size = torch.cuda.device_count()
    # world_size = 2
    # mp.spawn(run, args=(world_size, None, cfg), nprocs=world_size, join=True)
    mp.spawn(debug_random, args=(world_size, None, cfg), nprocs=world_size, join=True)
    # run(cfg)
    # Wandb.finish()


if __name__ == "__main__":
    main()



from .base import Distribution, NoMeanException
from .discrete import ConditionalIndependentBernoulli
from .normal import (
    ConditionalDiagonalNormal,
    DiagonalNormal,
    StandardNormal,
)
from .uniform import LotkaVolterraOscillating, MG1Uniform



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : main.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 10/23/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import tornado.ioloop

from jammy.logging import get_logger
from jamweb.web import make_app

logger = get_logger()


def main():
    app = make_app(
        ["app.index"],
        {
            "gzip": True,
            "debug": False,
            "xsrf_cookies": True,
            "static_path": "static",
            "template_path": "app/templates",
            "cookie_secret": "20f42d0ae6548e88cf9788e725b298bd",
            "session_secret": "3cdcb1f00803b6e78ab50b466a40b9977db396840c28307f428b25e2277f1bcc",
            "frontend_secret": "asdjikfh98234uf9pidwja09f9adsjp9fd6840c28307f428b25e2277f1bcc",
            "cookie_prefix": "jam_",
            # 'session_engine': 'off',
            "session_engine": "memcached",
            "session_cookie_prefix": "jam_sess_",
            "session_memcached_prefix": "jam_sess_",
            "session_timeout": 60 * 30,
            "memcached_host": "127.0.0.1",
            "memcached_port": "11211",
        },
    )
    app.listen(8081, xheaders=True)

    logger.critical("Mainloop started.")
    loop = tornado.ioloop.IOLoop.current()
    loop.start()


if __name__ == "__main__":
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
from typing import Optional, Type, Union

from .base import BaseStorageBackend
from .boto3_backend import Boto3Backend
from .http_backend import HTTPBackend
from .lmdb_backend import LmdbBackend
from .local_backend import LocalBackend
from .memcached_backend import MemcachedBackend

backends: dict = {}
prefix_to_backends: dict = {}


def _register_backend(
    name: str,
    backend: Type[BaseStorageBackend],
    force: bool = False,
    prefixes: Union[str, list, tuple, None] = None,
):
    """Register a backend.

    Args:
        name (str): The name of the registered backend.
        backend (BaseStorageBackend): The backend class to be registered,
            which must be a subclass of :class:`BaseStorageBackend`.
        force (bool): Whether to override the backend if the name has already
            been registered. Defaults to False.
        prefixes (str or list[str] or tuple[str], optional): The prefix
            of the registered storage backend. Defaults to None.
    """
    global backends, prefix_to_backends

    if not isinstance(name, str):
        raise TypeError("the backend name should be a string, " f"but got {type(name)}")

    if not inspect.isclass(backend):
        raise TypeError(f"backend should be a class, but got {type(backend)}")
    if not issubclass(backend, BaseStorageBackend):
        raise TypeError(f"backend {backend} is not a subclass of BaseStorageBackend")

    if name in backends and not force:
        raise ValueError(
            f"{name} is already registered as a storage backend, "
            'add "force=True" if you want to override it'
        )
    backends[name] = backend

    if prefixes is not None:
        if isinstance(prefixes, str):
            prefixes = [prefixes]
        else:
            assert isinstance(prefixes, (list, tuple))

        for prefix in prefixes:
            if prefix in prefix_to_backends and not force:
                raise ValueError(
                    f"{prefix} is already registered as a storage backend,"
                    ' add "force=True" if you want to override it'
                )

            prefix_to_backends[prefix] = backend


def register_backend(
    name: str,
    backend: Optional[Type[BaseStorageBackend]] = None,
    force: bool = False,
    prefixes: Union[str, list, tuple, None] = None,
):
    """Register a backend.

    Args:
        name (str): The name of the registered backend.
        backend (class, optional): The backend class to be registered,
            which must be a subclass of :class:`BaseStorageBackend`.
            When this method is used as a decorator, backend is None.
            Defaults to None.
        force (bool): Whether to override the backend if the name has already
            been registered. Defaults to False.
        prefixes (str or list[str] or tuple[str], optional): The prefix
            of the registered storage backend. Defaults to None.

    This method can be used as a normal method or a decorator.

    Examples:

        >>> class NewBackend(BaseStorageBackend):
        ...     def get(self, filepath):
        ...         return filepath
        ...
        ...     def get_text(self, filepath):
        ...         return filepath
        >>> register_backend('new', NewBackend)

        >>> @register_backend('new')
        ... class NewBackend(BaseStorageBackend):
        ...     def get(self, filepath):
        ...         return filepath
        ...
        ...     def get_text(self, filepath):
        ...         return filepath
    """
    if backend is not None:
        _register_backend(name, backend, force=force, prefixes=prefixes)
        return

    def _register(backend_cls):
        _register_backend(name, backend_cls, force=force, prefixes=prefixes)
        return backend_cls

    return _register


register_backend("local", LocalBackend, prefixes="")
register_backend("memcached", MemcachedBackend)
register_backend("lmdb", LmdbBackend)
# To avoid breaking backward Compatibility, 's3' is also used as a
# prefix for Boto3Backend
register_backend("s3", Boto3Backend, prefixes=["s3"])
register_backend("http", HTTPBackend, prefixes=["http", "https"])



# Copyright (c) OpenMMLab. All rights reserved.
from .base import BaseStorageBackend
from .boto3_backend import Boto3Backend
from .http_backend import HTTPBackend
from .lmdb_backend import LmdbBackend
from .local_backend import LocalBackend
from .memcached_backend import MemcachedBackend
from .registry_utils import backends, prefix_to_backends, register_backend

__all__ = [
    "BaseStorageBackend",
    "LocalBackend",
    "HTTPBackend",
    "LmdbBackend",
    "MemcachedBackend",
    "Boto3Backend",
    "register_backend",
    "backends",
    "prefix_to_backends",
]



import os
import os.path as osp
import time

import torch

from tqdm.auto import tqdm, trange
from jammy.cli.cmdline_viz import CmdLineViz
from jammy.event import SimpleEventRegistry
from jammy.logging import get_logger
from jammy.utils.enum import JamEnum
from jammy.utils.meter import AverageMeter
from jamtorch.utils.meta import as_float

from .monitor import group_prefix
from .trainer_monitor import TrainerMonitor
from .utils import *
from jammy.utils.hyd import *

logger = get_logger()

__all__ = ["Trainer"]


def cuda_time(sync=True):
    if sync:
        torch.cuda.synchronize()
    return time.time()


class EvalState(JamEnum):
    NO = 1
    ITER = 2
    EPOCH = 3


class Trainer:
    def __init__(self, model, optimizer, loss_fn, lr_scheduler=None, cfg=dict()):
        (self.model, self.optimizer, self.loss_fn, self.lr_scheduler,) = (
            model,
            optimizer,
            loss_fn,
            lr_scheduler,
        )
        self._event_manager = SimpleEventRegistry(
            {
                "epoch:start",
                "epoch:finish",
                "epoch:before",
                "epoch:after",
                "step:before",
                "step:after",
                "forward:before",
                "forward:after",
                "backward:before",
                "backward:after",
                "val:before",
                "val:after",
                "val:epoch",  # quite strange naming and design
            }
        )
        self.cmdviz = CmdLineViz()
        self.trainer_monitor = None

        # cache state accross different methods
        self.cur_monitor = dict()
        self.cur_cmdviz = dict()

        # trainer env states
        self._cfg = cfg
        self.load_env(cfg)

    def load_env(self, dict_cfg):
        self.eval_epoch = dict_cfg.get("eval_epoch") or 1
        self.eval_iter = dict_cfg.get("eval_iter") or -1
        self.best_loss = dict_cfg.get("best_loss") or float("inf")
        self.checkpoint_dir = dict_cfg.get("ckptDir") or os.getcwd()
        self.iter_cnt = dict_cfg.get("iter_cnt") or 0
        self.epoch_cnt = dict_cfg.get("epoch_cnt") or 0

    def export_env(self):
        return {
            "eval_epoch": self.eval_epoch,
            "eval_iter": self.eval_iter,
            "best_loss": self.best_loss,
            "ckptDir": self.checkpoint_dir,
            "iter_cnt": self.iter_cnt,
            "epoch_cnt": self.epoch_cnt,
        }

    def set_monitor(self, is_wandb, tblogger=False):
        """
        docstring
        """
        self.trainer_monitor = TrainerMonitor(is_wandb, tblogger)

    def load_ckpt(self, filename="checkpoint"):
        env = load_checkpoint(self.model, self.optimizer, hydpath(filename))
        self.load_env(env)

    def __call__(
        self,
        n_epochs,
        train_loader,
        val_loader,
        trainer_ckpt="checkpoint",
        is_resume=False,
    ):
        """
        resume training or start a new training
        """
        # FIXME
        if is_resume:
            self.load_ckpt(trainer_ckpt)
            start_epoch = self.epoch_cnt + 1
            start_iter = self.iter_cnt + 1
            loss = self.best_loss
        else:
            start_epoch, start_iter, loss = 0, 0, float("inf")
        return self.train(
            n_epochs, train_loader, val_loader, loss, start_iter, start_epoch
        )

    def eval_impl(self, data_loader, **kwargs):
        self.model.eval()
        self.trigger_event("val:before", self, data_loader)

        loss_meter = AverageMeter()
        with tqdm(total=len(data_loader), leave=False, desc="val") as pbar:
            with torch.no_grad():
                for i, batch in enumerate(data_loader):
                    # FIXME what is the usage of momitot, should be fix the api, make an agreement
                    loss, monitor, cmdviz_dict = self.loss_fn(
                        self.model, batch, is_train=False
                    )
                    # FIXME bug
                    self.trigger_event(
                        "val:epoch", self, batch, loss, monitor, cmdviz_dict
                    )
                    loss_meter.update(as_float(loss))
                    f_cmdviz = as_float(cmdviz_dict)
                    f_cmdviz.update({"loss": loss_meter.val})
                    self.cmdviz.update("eval", f_cmdviz)

                    pbar.update()
                    pbar.set_postfix({"loss": loss_meter.avg})

        self.trigger_event("val:after", self, data_loader)

        self.val_loss_ckpt(loss_meter.avg)
        self.cur_monitor.update(group_prefix("eval", self.cmdviz.meter["eval"].avg))

    def val_loss_ckpt(self, val_loss):
        is_best = val_loss < self.best_loss
        self.best_loss = min(val_loss, self.best_loss)
        if self.checkpoint_dir is not None:
            state = checkpoint_state(self.model, self.optimizer)
            state["env"] = self.export_env()

            save_checkpoint(
                state,
                is_best,
                osp.join(self.checkpoint_dir, "checkpoint"),
            )

    def loss_backward(self, loss):
        loss.backward()
        return True

    def train_step(self, feed_dict):
        self.trigger_event("step:before", self)
        self.trigger_event("forward:before", self, feed_dict)
        loss, monitors, cmdviz_dict = self.loss_fn(self.model, feed_dict, is_train=True)
        self.trigger_event(
            "forward:after", self, feed_dict, loss, monitors, cmdviz_dict
        )
        self.trigger_event(
            "backward:before", self, feed_dict, loss, monitors, cmdviz_dict
        )
        if loss.requires_grad:
            self.optimizer.zero_grad()
            self.trigger_event(
                "backward:before", self, feed_dict, loss, monitors, cmdviz_dict
            )
            # The design is useful for accumulated loss
            if self.loss_backward(loss):
                self.trigger_event(
                    "backward:after", self, feed_dict, loss, monitors, cmdviz_dict
                )
                self.optimizer.step()

        return self._train_step_after(loss, monitors, cmdviz_dict)

    def _train_step_after(self, loss, monitors, cmdviz_dict):
        self.trigger_event("step:after", self)

        loss_f = as_float(loss)
        cmdviz_f = as_float(cmdviz_dict)
        cmdviz_f.update({"loss": loss_f})

        monitors = as_float(monitors)
        monitors.update(cmdviz_f)

        self.cur_monitor.update(group_prefix("train", monitors))
        self.cmdviz.update("train", cmdviz_f)
        return loss_f

    def register_event(self, name, callback):
        """
        "epoch:start",(trainer)
        "epoch:finish",(trainer)
        "epoch:before",(trainer)
        "epoch:after",(trainer)
        "step:before",(trainer)
        "step:after",(trainer)
        "forward:before",(trainer, feed_dict)
        "forward:after",(trainer, feed_dict, loss, monitors, cmdviz)
        "backward:before",(trainer, feed_dict, loss, monitors, cmdviz)
        "backward:after",(trainer, feed_dict, loss, monitors, cmdviz)
        "val:before",(trainer, dataloader)
        "val:after",(trainer, dataloader)
        "val:epoch",(trainer, feed_dict, loss, monitors, cmdviz)
        """

        logger.info(
            "Register trainer event: name={}, callback={}.".format(
                name, callback.__module__ + "." + callback.__name__
            )
        )
        self._event_manager.register(name, callback)

    def trigger_event(self, name, *args, **kwargs):
        self._event_manager.trigger(name, *args, **kwargs)

    # TODO: strange API
    def train(
        self,
        n_epochs,
        train_loader,
        test_loader=None,
        best_loss=float("inf"),
        start_it=0,
        start_epoch=0,
    ):
        num_batch = len(train_loader)
        with trange(
            start_epoch, n_epochs, desc="epochs", dynamic_ncols=True
        ) as tbar, tqdm(
            total=num_batch, leave=False, desc="train", dynamic_ncols=True
        ) as pbar:
            self.trigger_event("epoch:start", self)
            for self.epoch_cnt in tbar:
                self.trigger_event("epoch:before", self)
                for batch_cnt, batch in enumerate(train_loader):
                    loss = self.train_step(batch)
                    self.iter_cnt += 1

                    pbar.update()
                    pbar.set_postfix(dict(total_it=self.iter_cnt, loss=loss))
                    state = self.eval_state(test_loader, batch_cnt, num_batch)
                    if state is EvalState.NO:
                        self.monitor_update()
                    elif state is EvalState.ITER:
                        pbar.close()
                        self.eval_impl(test_loader)
                        self.cmdviz.flush()
                        pbar = tqdm(
                            total=num_batch,
                            leave=False,
                            desc="train",
                            dynamic_ncols=True,
                            initial=batch_cnt + 1,
                        )
                        self.monitor_update()
                pbar.close()
                if state is EvalState.EPOCH:
                    self.eval_impl(test_loader)
                    self.cmdviz.flush()
                    self.monitor_update()

                self.trigger_event("epoch:after", self)
                pbar = tqdm(
                    total=num_batch,
                    leave=False,
                    desc="train",
                    dynamic_ncols=True,
                )

            self.trigger_event("epoch:finish", self)
        pbar.clear()
        return self.best_loss

    def eval_state(self, val_loader, batch_cnt, num_batch):
        if val_loader is None:
            return EvalState.NO
        if self.eval_iter > 0 and ((self.iter_cnt + 1) % self.eval_iter) == 0:
            if self.eval_epoch > 0 and ((self.epoch_cnt + 1) % self.eval_epoch) == 0:
                return EvalState.EPOCH
            return EvalState.ITER
        if batch_cnt + 1 == num_batch:
            if self.eval_epoch > 0 and ((self.epoch_cnt + 1) % self.eval_epoch) == 0:
                return EvalState.EPOCH
        return EvalState.NO

    def monitor_update(self):
        if self.trainer_monitor:
            self.trainer_monitor.update(
                {
                    **self.cur_monitor,
                    "epoch": self.epoch_cnt,
                    "iter": self.iter_cnt,
                }
            )
        self.cur_monitor = dict()



import torchvision
import torch
import jammy.utils.hyd as hyd
import os

def get_dataset():
    transform = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])
    train = torchvision.datasets.CIFAR10(root=hyd.hydpath('/tmp/data/train'), train=True, 
        download=True, transform=transform)
    val = torchvision.datasets.CIFAR10(root=hyd.hydpath('/tmp/data/val'), train=False, 
        download=True, transform=transform)
    return train, val



import os
import shutil
import socket

from omegaconf import OmegaConf

from jammy.utils import git

__all__ = ["Wandb", "WandbUrls"]


class WandbUrls:  # pylint: disable=too-few-public-methods
    def __init__(self, url):

        url_hash = url.split("/")[-1]
        project = url.split("/")[-3]
        entity = url.split("/")[-4]

        self.weight_url = url
        self.log_url = "https://app.wandb.ai/{}/{}/runs/{}/logs".format(
            entity, project, url_hash
        )
        self.chart_url = "https://app.wandb.ai/{}/{}/runs/{}".format(
            entity, project, url_hash
        )
        self.overview_url = "https://app.wandb.ai/{}/{}/runs/{}/overview".format(
            entity, project, url_hash
        )
        self.hydra_config_url = (
            "https://app.wandb.ai/{}/{}/runs/{}/files/hydra-config.yaml".format(
                entity, project, url_hash
            )
        )
        self.overrides_url = (
            "https://app.wandb.ai/{}/{}/runs/{}/files/overrides.yaml".format(
                entity, project, url_hash
            )
        )

    # pylint: disable=line-too-long
    def __repr__(self):
        msg = "=================================================== WANDB URLS ===================================================================\n"
        for k, v in self.__dict__.items():
            msg += "{}: {}\n".format(k.upper(), v)
        msg += "=================================================================================================================================\n"
        return msg

    def to_dict(self):
        return {k.upper(): v for k, v in self.__dict__.items()}


class Wandb:
    IS_ACTIVE = False
    IS_HYD = False
    cfg = None
    run = None

    @staticmethod
    def set_urls_to_model(model, url):
        wandb_urls = WandbUrls(url)
        model.wandb = wandb_urls

    @staticmethod
    def _set_to_wandb_args(wandb_args, cfg, name):
        var = getattr(cfg.wandb, name, None)
        if var:
            wandb_args[name] = var

    @staticmethod
    def check_repos():
        jam_sha, jam_diff = git.log_repo(__file__)
        from jammy.utils.env import jam_getenv

        if jam_getenv("proj_path"):
            proj_dir = jam_getenv("proj_path")
        else:
            import __main__ as _main

            proj_dir = _main.__file__
        project_sha, project_diff = git.log_repo(proj_dir)
        with open("jam_change.patch", "w", encoding="utf8") as f:
            f.write(jam_diff)
        with open("proj_change.patch", "w", encoding="utf8") as f:
            f.write(project_diff)

        return jam_sha, proj_dir, project_sha

    @staticmethod
    def prep_args(cfg):
        jam_sha, proj_dir, proj_sha = Wandb.check_repos()
        wandb_args = {
            "project": cfg.wandb.project,
            "resume": "allow",
            # "tags": cfg.wandb.tags,
            "config": {
                "run_path": os.getcwd(),
                "jam_sha": jam_sha,
                "proj_path": proj_dir,
                "proj_sha": proj_sha,
                "hydra": Wandb.IS_HYD,
                "host": socket.gethostname(),
            },
        }
        for key in ["name", "entity", "notes", "id", "tags"]:
            Wandb._set_to_wandb_args(wandb_args, cfg, key)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)
        if "wandb" in cfg_dict:
            del cfg_dict["wandb"]

        wandb_args["config"]["z"] = cfg_dict

        return wandb_args

    @staticmethod
    def launch(cfg, launch: bool, is_hydra: bool = True, dump_meta: bool = True):
        Wandb.IS_HYD = is_hydra
        if launch:
            import wandb

            Wandb.IS_ACTIVE = True
            wandb_args = Wandb.prep_args(cfg)
            Wandb.run = wandb.init(**wandb_args)
            Wandb.cfg = {**wandb_args["config"], **(WandbUrls(Wandb.run.url).to_dict())}

            wandb.save(os.path.join(os.getcwd(), "jam_change.patch"))
            wandb.save(os.path.join(os.getcwd(), "proj_change.patch"))

            if is_hydra:
                shutil.copyfile(
                    os.path.join(os.getcwd(), ".hydra/config.yaml"),
                    os.path.join(os.getcwd(), ".hydra/hydra-config.yaml"),
                )
                wandb.save(os.path.join(os.getcwd(), ".hydra/hydra-config.yaml"))
                wandb.save(os.path.join(os.getcwd(), ".hydra/overrides.yaml"))
        else:
            Wandb.cfg = Wandb.prep_args(cfg)["config"]
        if dump_meta:
            with open("meta.yaml", "w", encoding="utf8") as fp:
                OmegaConf.save(config=OmegaConf.create(Wandb.cfg), f=fp.name)
        return Wandb.run

    @staticmethod
    def add_file(file_path: str):
        if not Wandb.IS_ACTIVE:
            return
        import wandb

        filename = os.path.basename(file_path)
        shutil.copyfile(file_path, os.path.join(wandb.run.dir, filename))

    @staticmethod
    def log(*args, **kargs):
        if not Wandb.IS_ACTIVE:
            raise RuntimeError("wandb is inactive, please launch first.")
        import wandb

        wandb.log(*args, **kargs)

    @staticmethod
    def finish():
        if not Wandb.IS_ACTIVE:
            return
        import wandb

        if os.path.exists("jam_.log"):
            Wandb.add_file("jam_.log")

        wandb.finish()



from collections import OrderedDict
from dataclasses import asdict, dataclass
from typing import Any, Callable, List, Optional, Tuple, TypeVar

import pandas as pd
import torch
from torch import Tensor, nn
from torch.utils.hooks import RemovableHandle

DataclassType = TypeVar("DataclassType", bound=Any)  # pylint: disable=invalid-name
ProcessFn = Callable[[str, nn.Module, int, Tensor], DataclassType]


@dataclass
class TeorKStates:
    name: str
    shape: Tuple[int, ...]
    dtype: str
    magnitude: float
    max: float
    viz_type: Optional[str] = None


def terok_processor(
    name: str, module: nn.Module, idx: int, tensor: Tensor
) -> TeorKStates:
    del module, idx
    return TeorKStates(
        name=name,
        shape=tensor.shape,
        dtype=str(tensor.dtype).rsplit(".", maxsplit=1)[-1],
        magnitude=float(tensor.to(torch.float32).square().mean().sqrt()),
        max=float(tensor.to(torch.float32).abs().max()),
    )


class PerLayerSpec:
    def __init__(
        self,
        process_fn: ProcessFn = terok_processor,
    ):
        self.module2name = OrderedDict()
        self.unique_names = set()

        self.layer_specs = []
        self.process_fn = process_fn

    @property
    def results(self) -> List[DataclassType]:
        return self.layer_specs

    def unique_name(self, module, idx, num):
        name = self.module2name[module]

        # If there is more than one instance of the module, append the index to its name.
        # This helps in differentiating multiple instances of the same module.
        if num > 1:
            name += f":{idx}"

        # Check if the constructed name is already used by another module instance.
        # If it is, a unique suffix needs to be added to the name.
        if name in self.unique_names:
            # Start with a suffix of 2 (assuming the original without suffix is considered '1').
            suffix = 2
            # Increment the suffix until a unique name is found that isn't in the set of unique names.
            while f"{name}_{suffix}" in self.unique_names:
                suffix += 1
            # Once a unique suffix is found, append it to the name.
            name += f"_{suffix}"

        # Add the newly created unique name to the set of unique names to keep track.
        self.unique_names.add(name)

        # Return the unique name for the module.
        return name

    def module_hook(self, module, _, outputs):
        outputs = list(outputs) if isinstance(outputs, (tuple, list)) else [outputs]
        outputs = [out for out in outputs if isinstance(out, torch.Tensor)]
        for idx, value in enumerate(outputs):
            name = self.unique_name(module, idx, len(outputs))
            self.layer_specs.append(self.process_fn(name, module, idx, value))

    def register_hook(self, name: str, module: nn.Module) -> RemovableHandle:
        self.module2name[module] = name
        return module.register_forward_hook(self.module_hook)

    def reset(self):
        self.module2name.clear()
        self.unique_names.clear()
        self.layer_specs.clear()

    def reset_forward(self):
        self.layer_specs.clear()
        self.unique_names.clear()


def layer_sepc_to_df(layer_specs: List[DataclassType]) -> pd.DataFrame:
    return pd.DataFrame([asdict(spec) for spec in layer_specs])



import glob
from PIL import Image

from jammy.cli.argument import JamArgumentParser

parser = JamArgumentParser()
parser.add_argument("--fmt", required=True, help="glob foramt eg: 'img_*.png'")
parser.add_argument("--out", default="output.gif")
args = parser.parse_args()

# TODO: consider other implementations, https://github.com/wkentaro/video-cli

# https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
def fp2gif(fp_in, fp_out):
    img, *imgs = [Image.open(f) for f in sorted(glob.glob(fp_in))]
    img.save(
        fp=fp_out, format="GIF", append_images=imgs, save_all=True, duration=200, loop=0
    )


if __name__ == "__main__":
    fp2gif(args.fmt, args.out)



from jammy.utils.cnt import CBCnt, ExitCnt


def test_cbcnt():
    N = 10
    cnter = CBCnt(lambda: print("Callback starts running"), N)
    for i in range(3 * N):
        cnter(i % 2)
    for _ in range(3 * N):
        cnter(True)


def test_exitcnt():
    N = 10
    cnter = ExitCnt("Quit", N)
    for _ in range(3 * N):
        cnter(True)


if __name__ == "__main__":
    test_cbcnt()
    test_exitcnt()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : lmdb.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.
import os
import pickle
from typing import Any, Iterable, Optional

from jammy.logging import get_logger
from jammy.utils.cache import cached_property
from jammy.utils.container import OrderedSet

from .kv import KVStoreBase

__all__ = ["LMDBKVStore"]

logger = get_logger()

_loads = pickle.loads
_dumps = pickle.dumps

# pylint: disable=attribute-defined-outside-init
class LMDBKVStore(KVStoreBase):
    """A LMDB-based key-value store. This class also supports sub-databases.


    Basic usage with a single database:
        .. code-block:: python

            kv = LMDBKVStore('/tmp/test_1.lmdb', readonly=False)

            with kv.transaction():
                kv['a'] = 1
                kv['b'] = 2

            assert 'a' in kv and kv['a'] == 1
            assert 'b' in kv and kv['b'] == 2
            assert 'c' not in kv

            for k in kv.keys():
                print(k, kv[k])
    """

    _key_charset = "utf8"
    _magic_key = b"__keys__"
    _magic_main_databse = b"__main_database__"

    def __init__(
        self,
        lmdb_path: str,
        readonly: bool = True,
        max_dbs: int = 0,
        keys: Optional[Iterable[Any]] = None,
    ):
        """Initialize the LMDBKVStore.

        Args:
            lmdb_path: the path to the LMDB file. By default, this path is a directory.
            readonly: whether to open the LMDB in readonly mode.
            max_dbs: the maximum number of sub-databases. 0 means the single-database mode.
            keys: the keys in the main database. For new databases, use None.
        """
        super().__init__(readonly=readonly)
        self._max_dbs = max_dbs
        self._open(lmdb_path, readonly=readonly, keys=keys, max_dbs=max_dbs)

    def _open(self, lmdb_path, readonly, keys, max_dbs):
        try:
            import lmdb  # noqa: F401 # pylint: disable=import-outside-toplevel
        except ImportError:
            raise ImportError('Please run "pip install lmdb" to enable LmdbBackend.')
        self._lmdb = lmdb.open(
            lmdb_path,
            subdir=os.path.isdir(lmdb_path) or not os.path.exists(lmdb_path),
            readonly=readonly,
            lock=False,
            readahead=False,
            map_size=1099511627776 * 2,
            max_readers=100,
            max_dbs=max_dbs,
        )
        self._subdbs = {}
        self._lmdb_keys = {}
        self._is_dirty = {}

        if keys is not None:
            self._lmdb_keys[self._magic_main_databse] = OrderedSet(keys)
            self._is_dirty[self._magic_main_databse] = True

        self._is_in_transaction = False

    @cached_property
    def txn(self):
        """The current LMDB transaction."""
        return self._lmdb.begin(write=not self.readonly)

    @property
    def lmdb(self):
        """The LMDB environment."""
        return self._lmdb

    def get_subdb(self, name=None):
        """Get a sub-database by name."""
        if name is None:
            return None
        if name not in self._subdbs:
            self._subdbs[name] = self._lmdb.open_db(
                name.encode(self._key_charset), txn=self.txn
            )
        return self._subdbs[name]

    def set_dirty(self, db=None):
        """Set the dirty flag for the given sub-database."""
        self._is_dirty[db] = True

    def _has(self, key, db=None):
        self._load_lmdb_keys(db)
        return key in self._lmdb_keys[db]

    def _get(self, key, default, db=None):
        value = self.txn.get(
            key.encode(self._key_charset), default=default, db=self.get_subdb(db)
        )
        value = _loads(value)
        return value

    def _put(self, key, value, replace=False, db=None):
        assert (
            self._is_in_transaction
        ), "LMDBKVStore does not support put outside transaction."
        self._load_lmdb_keys(db)
        self._is_dirty[db] = True
        self._lmdb_keys[db].append(key)
        return self.txn.put(
            key.encode(self._key_charset),
            _dumps(value),
            overwrite=replace,
            db=self.get_subdb(db),
        )

    def _erase(self, key, db=None):
        assert (
            self._is_in_transaction
        ), "LMDBKVStore does not support erase outside transaction."
        self._load_lmdb_keys(db)
        self._is_dirty[db] = True
        # self._lmdb_keys.remove(key)
        del self._lmdb_keys[key]
        return self.txn.pop(key.encode(self._key_charset), db=self.get_subdb(db))

    def _keys(self, db=None):
        self._load_lmdb_keys(db)
        return self._lmdb_keys[db]

    def _transaction(self, *args, **kwargs):
        return self

    def __enter__(self):
        self._is_in_transaction = True
        return self

    def __exit__(self, exc_type, exc_value, exc_trace):
        self._is_in_transaction = False
        if exc_type:
            self.txn.abort()
        else:
            for k, v in self._is_dirty.items():
                if v:
                    self.txn.put(
                        self._magic_key,
                        _dumps(self._lmdb_keys[k].as_list()),
                        db=self.get_subdb(k),
                    )
                    self._is_dirty[k] = False
            self.txn.commit()

        self.txn = self._lmdb.begin(write=not self.readonly)

    def _load_lmdb_keys(self, db=None, assert_exist=False):
        if db in self._lmdb_keys:
            return
        dbobj = self.get_subdb(db)

        keys = self.txn.get(self._magic_key, None, db=dbobj)
        if assert_exist:
            assert keys is not None, "LMDBKVStore does not support __keys__ access"

        if keys is None:
            # build the list of keys by traversing the databse
            keys = OrderedSet()
            with self.txn.cursor(db=dbobj) as cursor:
                for k, _ in cursor:
                    keys.append(k.decode(self._key_charset))
            self._lmdb_keys[db] = keys
            self._is_dirty[db] = True

            if db is None and self._max_dbs > 0 and len(keys) > 0:
                logger.warning(
                    "LMDBKVStore is trying to build a list of keys of an non-empty main database. "
                    "This means that the underlying lmdb databse was not created by this class. "
                    "Because LMDB internally stores sub databases as links in the main database, this operation will"
                    "create a union of all keys in the main database and the names of the sub databases. "
                    "In most cases this is not what you want."
                    "You can use kv._lmdb_keys[None].remove() to remove the names of the sub databases, "
                    "and then call kv.set_dirty() to trigger auto-commit."
                )
        else:
            self._lmdb_keys[db] = OrderedSet(_loads(keys))



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod


class BaseFileHandler(metaclass=ABCMeta):
    # `str_like` is a flag to indicate whether the type of file object is
    # str-like object or bytes-like object. Pickle only processes bytes-like
    # objects but json only processes str-like object. If it is str-like
    # object, `StringIO` will be used to process the buffer.
    str_like = True

    @abstractmethod
    def load_from_fileobj(self, file, **kwargs):
        pass

    @abstractmethod
    def dump_to_fileobj(self, obj, file, **kwargs):
        pass

    @abstractmethod
    def dump_to_str(self, obj, **kwargs):
        pass

    def load_from_path(self, filepath, mode="r", **kwargs):
        with open(filepath, mode) as f:
            return self.load_from_fileobj(f, **kwargs)

    def dump_to_path(self, obj, filepath, mode="w", **kwargs):
        with open(filepath, mode) as f:
            self.dump_to_fileobj(obj, f, **kwargs)



"""Library for monitoring PyTorch network statistics.

This module provides utilities for monitoring various statistics of PyTorch networks,
including weights, gradients, and activations. It helps identify potential issues
like vanishing/exploding gradients and unusual distributions.
"""

import math
from dataclasses import dataclass
from typing import Optional

import numpy as np
import torch
from torch import Tensor


@dataclass
class TensorQuartileInfo:
    """Stores quartile information of a tensor distribution.

    Attributes:
        quartile_95: The 95th percentile value.
        quartile_75: The 75th percentile value (upper quartile).
        quartile_25: The 25th percentile value (lower quartile).
        quartile_5: The 5th percentile value.
    """

    quartile_95: float
    quartile_75: float
    quartile_25: float
    quartile_5: float

    def __repr__(self) -> str:
        """Returns a formatted string representation of quartile information."""
        return (
            f"TensorQuartileInfo(\n"
            f"    q95: {self.quartile_95:8.3g},\n"
            f"    q75: {self.quartile_75:8.3g},\n"
            f"    q25: {self.quartile_25:8.3g},\n"
            f"    q05: {self.quartile_5:8.3g}\n"
            f")"
        )

    @classmethod
    def from_tensor(cls, tensor: Tensor) -> "TensorQuartileInfo":
        """Creates TensorQuartileInfo from a tensor.

        Args:
            tensor: Input tensor of any shape, will be flattened for analysis.

        Returns:
            TensorQuartileInfo containing the quartile statistics of the input tensor.
        """
        flat_tensor = tensor.detach().float().flatten()
        percentiles = torch.tensor([5, 25, 75, 95], dtype=torch.float32)
        q05, q25, q75, q95 = torch.quantile(flat_tensor, percentiles / 100)

        return cls(
            quartile_95=q95.item(),
            quartile_75=q75.item(),
            quartile_25=q25.item(),
            quartile_5=q05.item(),
        )


@dataclass
class TensorOutlierInfo:
    """Stores outlier information using multiple detection methods.

    Attributes:
        iqr_outlier_ratio_high: Ratio of high outliers using IQR method.
        iqr_outlier_ratio_low: Ratio of low outliers using IQR method.
        modified_z_score_outlier_ratio_high: Ratio of high outliers using modified Z-score.
        modified_z_score_outlier_ratio_low: Ratio of low outliers using modified Z-score.
        mad_outlier_ratio_high: Ratio of high outliers using MAD method.
        mad_outlier_ratio_low: Ratio of low outliers using MAD method.
    """

    iqr_outlier_ratio_high: float
    iqr_outlier_ratio_low: float
    modified_z_score_outlier_ratio_high: float
    modified_z_score_outlier_ratio_low: float
    mad_outlier_ratio_high: float
    mad_outlier_ratio_low: float

    def __repr__(self) -> str:
        """Returns a formatted string representation of outlier information."""
        return (
            f"TensorOutlierInfo(\n"
            f"    IQR High:      {self.iqr_outlier_ratio_high:8.3%}\n"
            f"    IQR Low:       {self.iqr_outlier_ratio_low:8.3%}\n"
            f"    Mod Z High:    {self.modified_z_score_outlier_ratio_high:8.3%}\n"
            f"    Mod Z Low:     {self.modified_z_score_outlier_ratio_low:8.3%}\n"
            f"    MAD High:      {self.mad_outlier_ratio_high:8.3%}\n"
            f"    MAD Low:       {self.mad_outlier_ratio_low:8.3%}\n"
            f")"
        )

    @classmethod
    def from_tensor(cls, tensor: Tensor) -> "TensorOutlierInfo":
        """Creates TensorOutlierInfo from a tensor using multiple outlier detection methods.

        Args:
            tensor: Input tensor of any shape, will be flattened for analysis.

        Returns:
            TensorOutlierInfo containing outlier statistics using various methods.
        """
        flat_tensor = tensor.detach().float().flatten().numpy()

        # IQR method
        q75, q25 = np.percentile(flat_tensor, [75, 25])
        iqr = q75 - q25
        iqr_high = q75 + 1.5 * iqr
        iqr_low = q25 - 1.5 * iqr

        # Modified Z-score method
        median = np.median(flat_tensor)
        mad = np.median(np.abs(flat_tensor - median))
        modified_z = 0.6745 * (flat_tensor - median) / mad
        z_threshold = 3.5

        # MAD method
        mad_high = median + 3 * mad
        mad_low = median - 3 * mad

        return cls(
            iqr_outlier_ratio_high=np.mean(flat_tensor > iqr_high),
            iqr_outlier_ratio_low=np.mean(flat_tensor < iqr_low),
            modified_z_score_outlier_ratio_high=np.mean(modified_z > z_threshold),
            modified_z_score_outlier_ratio_low=np.mean(modified_z < -z_threshold),
            mad_outlier_ratio_high=np.mean(flat_tensor > mad_high),
            mad_outlier_ratio_low=np.mean(flat_tensor < mad_low),
        )


@dataclass
class TensorStatInfo:  # pylint: disable=too-many-instance-attributes
    """Comprehensive statistical information about a tensor.

    Attributes:
        num_samples: Number of elements in the tensor.
        m1: First moment (mean).
        m2: Second moment.
        median: Median value.
        std: Standard deviation.
        min: Minimum value.
        max: Maximum value.
        quartile_info: Quartile statistics.
        outlier_info: Outlier statistics.
    """

    num_samples: int
    m1: float
    m2: float
    median: float
    std: float
    min: float
    max: float
    quartile_info: TensorQuartileInfo
    outlier_info: TensorOutlierInfo

    def __repr__(self) -> str:
        """Returns a formatted string representation of all statistics."""
        return (
            f"StatInfo(\n"
            f"    num_samples: {self.num_samples}\n"
            f"    mean:        {self.m1:8.3g}\n"
            f"    std:         {self.std:8.3g}\n"
            f"    median:      {self.median:8.3g}\n"
            f"    min:         {self.min:8.3g}\n"
            f"    max:         {self.max:8.3g}\n"
            f"    quartile_info:\n"
            f"{self._indent(self.quartile_info.__repr__())}\n"
            f"    outlier_info:\n"
            f"{self._indent(self.outlier_info.__repr__())}\n"
            f")"
        )

    @staticmethod
    def _indent(text: str, spaces: int = 8) -> str:
        """Indents each line of the text by specified number of spaces."""
        return "\n".join(" " * spaces + line for line in text.split("\n"))

    @classmethod
    def from_tensor(cls, tensor: Tensor) -> "TensorStatInfo":
        """Creates StatInfo from a tensor.

        Args:
            tensor: Input tensor of any shape, will be flattened for analysis.

        Returns:
            StatInfo containing comprehensive statistics of the input tensor.
        """
        flat_tensor = tensor.detach().float().flatten()

        return cls(
            num_samples=flat_tensor.numel(),
            m1=torch.mean(flat_tensor).item(),
            m2=torch.mean(flat_tensor**2).item(),
            median=torch.median(flat_tensor).item(),
            std=torch.std(flat_tensor).item(),
            min=torch.min(flat_tensor).item(),
            max=torch.max(flat_tensor).item(),
            quartile_info=TensorQuartileInfo.from_tensor(flat_tensor),
            outlier_info=TensorOutlierInfo.from_tensor(flat_tensor),
        )


@dataclass
class TensorSparseSample:
    """Stores sparse samples from multiple tensors using indices and values.

    This class enables efficient storage and updates of specific tensor elements,
    useful for monitoring large tensors without storing the entire tensor.
    It can track values from multiple tensor updates, stacking them for analysis.

    Attributes:
        indices: Tensor of indices for the sampled elements. Shape: (n_samples, n_dims)
            For 1D tensors, shape is (n_samples,).
        values: Optional tensor of values at the specified indices.
            Shape is (ntensor, n_samples) where ntensor is number of updates.
        ntensor: Number of tensors processed (number of updates made).
    """

    indices: Tensor
    values: Optional[Tensor] = None
    ntensor: int = 0

    def __post_init__(self):
        """Validates the inputs after initialization.

        Raises:
            TypeError: If indices or values are not tensors.
            ValueError: If values shape doesn't match indices or has wrong dimensions.
        """
        if not isinstance(self.indices, Tensor):
            raise TypeError("indices must be a torch.Tensor")

        if self.values is not None:
            if not isinstance(self.values, Tensor):
                raise TypeError("values must be a torch.Tensor")

            # Check values is 2D
            if self.values.dim() != 2:
                raise ValueError(
                    "values must be a 2D tensor with shape (ntensor, n_samples)"
                )

            # Check number of samples matches
            if self.values.shape[1] != self.indices.shape[0]:
                raise ValueError(
                    f"Number of values ({self.values.shape[1]}) must match "
                    f"number of indices ({self.indices.shape[0]})"
                )

            # Update ntensor based on values if provided
            self.ntensor = self.values.shape[0]

    def update(self, tensor: Tensor) -> None:
        """Updates values from the input tensor at stored indices.

        This method extracts values at the stored indices and stacks them with
        previous updates. Each update adds a new row to the values tensor.

        Args:
            tensor: Input tensor to sample values from. Its shape must be compatible
                with the stored indices.

        Raises:
            ValueError: If tensor shape is incompatible with stored indices.
            RuntimeError: If indices are out of bounds for the input tensor.
        """
        if not isinstance(tensor, Tensor):
            raise TypeError("tensor must be a torch.Tensor")

        # Handle different dimensional cases
        if self.indices.dim() == 1:
            if tensor.dim() != 1:
                raise ValueError(
                    f"Expected 1D tensor for 1D indices, got shape {tensor.shape}"
                )
            new_values = tensor[self.indices]
        else:
            # For multi-dimensional tensors
            try:
                new_values = tensor[tuple(self.indices.t())]
            except IndexError as e:
                raise RuntimeError(
                    f"Index out of bounds. Tensor shape: {tensor.shape}, "
                    f"Max index: {self.indices.max().item()}"
                ) from e

        new_values = new_values.detach().clone()

        # Stack with previous values or initialize
        if self.values is None:
            self.values = new_values.unsqueeze(0)
        else:
            self.values = torch.cat([self.values, new_values.unsqueeze(0)])

        self.ntensor += 1

    @classmethod
    def from_tensor(
        cls,
        tensor: Tensor,
        indices: Optional[Tensor] = None,
        sample_size: Optional[int] = None,
        strategy: str = "random",
    ) -> "TensorSparseSample":
        """Creates a TensorSparseSample from a tensor.

        Args:
            tensor: Input tensor to sample from.
            indices: Optional specific indices to sample. If None, indices will be
                generated according to the strategy.
            sample_size: Number of samples to take if indices is None.
                Defaults to min(sqrt(tensor.numel()), 100) if None.
            strategy: Sampling strategy when generating indices. Options:
                - 'random': Random sampling (default)
                - 'uniform': Uniform sampling across the tensor
                - 'edges': Sample from edges and center of the tensor

        Returns:
            A new TensorSparseSample instance containing the sampled indices and values.

        Raises:
            ValueError: If strategy is invalid or sampling parameters are inconsistent.
        """
        if not isinstance(tensor, Tensor):
            raise TypeError("tensor must be a torch.Tensor")

        # Generate indices if not provided
        if indices is None:
            # Determine sample size if not specified
            if sample_size is None:
                sample_size = min(int(math.sqrt(tensor.numel())), 100)

            indices = cls._generate_indices(tensor, sample_size, strategy)
        else:
            if not isinstance(indices, Tensor):
                raise TypeError("indices must be a torch.Tensor")

            # Validate indices
            if indices.max() >= tensor.numel():
                raise ValueError("Index out of bounds for tensor")

        # Create instance and update values
        instance = cls(indices=indices, ntensor=0)
        instance.update(tensor)
        return instance

    @staticmethod
    def _generate_indices(  # pylint: disable=too-many-locals
        tensor: Tensor, sample_size: int, strategy: str = "random"
    ) -> Tensor:
        """Generates indices for sampling based on the specified strategy.

        Args:
            tensor: Input tensor to generate indices for.
            sample_size: Number of samples to generate.
            strategy: Sampling strategy ('random', 'uniform', or 'edges').

        Returns:
            Tensor containing the generated indices.

        Raises:
            ValueError: If strategy is invalid or sample_size is too large.
        """
        if sample_size > tensor.numel():
            raise ValueError(
                f"sample_size ({sample_size}) cannot be larger than "
                f"number of elements ({tensor.numel()})"
            )

        # Generate linear indices first
        if strategy == "random":
            linear_indices = torch.randperm(tensor.numel())[:sample_size]

        elif strategy == "uniform":
            stride = tensor.numel() // sample_size
            linear_indices = torch.arange(0, tensor.numel(), stride)[:sample_size]

        elif strategy == "edges":
            # Take samples from start, middle, and end of tensor
            third = sample_size // 3
            remainder = sample_size % 3

            start_indices = torch.arange(third)
            middle_start = (tensor.numel() // 2) - (third // 2)
            middle_indices = torch.arange(middle_start, middle_start + third)
            end_start = tensor.numel() - third - remainder
            end_indices = torch.arange(end_start, tensor.numel())

            linear_indices = torch.cat([start_indices, middle_indices, end_indices])

        else:
            raise ValueError(
                f"Unknown sampling strategy: {strategy}. "
                "Valid options are: 'random', 'uniform', 'edges'"
            )

        # For multi-dimensional tensors, convert linear indices to multi-dimensional indices
        if tensor.dim() > 1:
            multi_indices = []
            shape = tensor.shape
            for idx in linear_indices:
                # Convert linear index to multi-dimensional indices
                curr_idx = idx.item()
                multi_idx = []
                for dim_size in shape[::-1]:
                    multi_idx.append(curr_idx % dim_size)
                    curr_idx //= dim_size
                multi_indices.append(multi_idx[::-1])  # Reverse to get correct order
            return torch.tensor(multi_indices)

        return linear_indices

    def __len__(self) -> int:
        """Returns the number of samples."""
        return self.indices.shape[0]

    def __repr__(self) -> str:
        """Returns a string representation of the sparse sample."""
        return (
            f"TensorSparseSample(\n"
            f"    num_samples: {len(self)}\n"
            f"    indices shape: {tuple(self.indices.shape)}\n"
            f"    values: {'Not set' if self.values is None else f'shape {tuple(self.values.shape)}'}\n"
            f"    ntensor: {self.ntensor}\n"
            ")"
        )


def get_tensor_stat_info(tensor: Tensor, name: Optional[str] = None) -> TensorStatInfo:
    """Returns comprehensive statistical information about a tensor.

    Args:
        tensor: Input tensor of any shape.
        name: Optional name to identify the tensor.

    Returns:
        StatInfo containing comprehensive statistics of the input tensor.
    """
    if name is not None:
        print(f"Statistics for {name} tensor:")
    return TensorStatInfo.from_tensor(tensor)



try:
    from apex import amp

    APEX_AVAILABLE = True
except:
    APEX_AVAILABLE = False

import torch

from .genetic_trainer import GeneticTrainer


__all__ = ["loss_backwards", "fp16_wrapper"]


def loss_backwards(fp16, loss, optimizer, **kwargs):
    if fp16:
        with amp.scale_loss(loss, optimizer) as scaled_loss:
            scaled_loss.backward(**kwargs)
    else:
        loss.backward(**kwargs)


def amp_init(trainer, *args, **kwargs):
    trainer.model, trainer.optimizer = amp.initialize(
        trainer.model, trainer.optimizier, opt_level="O1"
    )


def amp_state_dict(*args, **kwargs):
    return {"amp": amp.state_dict()}


def amp_load_state(state, *args, **kwargs):
    """
    call after the initialization!!! Pay attention to the location.
    """
    amp.load_state_dict(state["amp"])

# Replace the GeneticTrainer by any trainer
class Fp16GeneticTrainer(GeneticTrainer):
    def __init__(self, cfg, loss_fn):
        super().__init__(cfg, loss_fn)
        
        self.fp16 = cfg.get("fp16") or False
        self.setup_fp16()

    def set_model_optim(self, model, optimizer):
        if self.fp16:
            model, optimizer = amp.initialize(
                    model, optimizer, opt_level="O1"
                )
        super().set_model_optim(model, optimizer=optimizer)

    def loss_backward(self, loss):
        if self.fp16:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                value = super().loss_backward(scaled_loss)
            return value
        return super().loss_backward(loss)

    def setup_fp16(self):
        def _fp16_load(trainer, state):
            amp_load_state(state)
        def _fp16_export(trainer, state_dict):
            state_dict.update(amp_state_dict())
        self.register_event("trainer:load",_fp16_load)
        self.register_event("trainer:export", _fp16_export)


def fp16_wrapper(base_class):
    class _FPClass(base_class):
        def __init__(self,*args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

            self.fp16 = self._cfg.get("fp16") or False
            self.setup_fp16()

        def set_model_optim(self, model, optimizer):
            if next(model.parameters()).device == torch.device("cpu"):
                model = model.to(self.device)
            if self.fp16:
                model, optimizer = amp.initialize(
                        model, optimizer, opt_level="O1"
                    )
            super().set_model_optim(model, optimizer=optimizer)

        def loss_backward(self, loss):
            if self.fp16:
                with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                    value = super().loss_backward(scaled_loss)
                return value
            return super().loss_backward(loss)

        def setup_fp16(self):
            def _fp16_load(trainer, state):
                if trainer.fp16:
                    amp_load_state(state)
            def _fp16_export(trainer, state_dict):
                if trainer.fp16:
                    state_dict.update(amp_state_dict())
            self.register_event("trainer:load",_fp16_load)
            self.register_event("trainer:export", _fp16_export)

    return _FPClass


import gpustat
import numpy as np


def is_gpu_free(ids=0, mem_thres=0.2, util_thres=0.3):
    if isinstance(ids, (list, tuple)):
        return all(is_gpu_free(item, mem_thres, util_thres) for item in ids)
    if isinstance(ids, int):
        query = gpustat.new_query()
        fixed_gpu = query[ids]
        used_mem = 1.0 * fixed_gpu.memory_used / fixed_gpu.memory_total
        if used_mem < mem_thres and fixed_gpu.utilization < util_thres:
            return True
        return False
    if ids == "all":
        query = gpustat.new_query()
        return is_gpu_free(list(range(len(query))), mem_thres, util_thres)
    raise RuntimeError(f"{ids} not supprted")


def gpu_by_weight(mem_prior=1.0):
    mem_prior = np.clip(mem_prior, 0.0, 1.0)
    query = gpustat.new_query()
    if len(query) == 0:
        raise RuntimeError("gpu not available")
    if len(query) == 1:
        return query[0].entry["index"]

    mem_list, utils_list = get_mem_util()
    mem, utils = np.array(mem_list), np.array(utils_list)
    weight = mem * mem_prior + utils * (1 - mem_prior)

    ids = np.argsort(weight)
    return [query[id_item].entry["index"] for id_item in ids]


def gpu_by_util():
    query = gpustat.new_query()
    _, utils_list = get_mem_util()
    ids = np.argsort(utils_list)
    return [query[id_item].entry["index"] for id_item in ids]


def get_mem_util():
    query = gpustat.new_query()
    used_space_list = [1.0 * item.memory_used / item.memory_total for item in query]
    utils_list = [item.utilization for item in query]
    return used_space_list, utils_list



# pylint: disable=import-outside-toplevel,import-error,redefined-builtin,too-many-arguments
from typing import IO

import numpy as np

from .base import BaseFileHandler

try:
    import imageio
except ImportError:
    imageio = None


class ImageioVideoHandler(BaseFileHandler):
    str_like = False

    def load_from_fileobj(
        self, file: IO[bytes], format: str = "mp4", mode: str = "rgb", **kwargs
    ):
        """
        Load video from a file-like object using imageio with specified format and color mode.

        Parameters:
            file (IO[bytes]): A file-like object containing video data.
            format (str): Format of the video file (default 'mp4').
            mode (str): Color mode of the video, 'rgb' or 'gray' (default 'rgb').

        Returns:
            tuple: A tuple containing an array of video frames and metadata about the video.
        """
        file.seek(0)
        video_reader = imageio.get_reader(file, format, **kwargs)

        video_frames = []
        for frame in video_reader:
            if mode == "gray":
                import cv2  # Convert frame to grayscale if mode is gray

                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
                frame = np.expand_dims(
                    frame, axis=2
                )  # Keep frame dimensions consistent
            video_frames.append(frame)

        return np.array(video_frames), video_reader.get_meta_data()

    def dump_to_fileobj(
        self,
        obj: np.ndarray,
        file: IO[bytes],
        format: str = "mp4",  # pylint: disable=redefined-builtin
        fps: int = 30,
        quality: int = 5,
        **kwargs,
    ):
        """
        Save an array of video frames to a file-like object using imageio.

        Parameters:
            obj (np.ndarray): An array of frames to be saved as video.
            file (IO[bytes]): A file-like object to which the video data will be written.
            format (str): Format of the video file (default 'mp4').
            fps (int): Frames per second of the output video (default 30).

        """
        # with imageio.get_writer(file, format=format, fps=fps, **kwargs) as writer:
        #     for frame in obj:
        #         writer.append_data(frame)
        h, w = obj.shape[1:-1]
        kwargs = {
            "fps": fps,
            "quality": quality,
            "macro_block_size": 1,
            "ffmpeg_params": ["-s", f"{w}x{h}"],
            "output_params": ["-f", "mp4"],
        }
        imageio.mimsave(file, obj, format, **kwargs)

    def dump_to_str(self, obj, **kwargs):
        raise NotImplementedError



import numpy as np
from torch import nn

__all__ = ["BatchNorm1d","BatchNorm2d","BatchNorm3d","Conv1d",\
    "Conv2d","Conv3d","FC"]

class BaseModule(nn.Module):
    """ Base module class with some basic additions to the pytorch Module class
    """

    @property
    def nb_params(self):
        """This property is used to return the number of trainable parameters for a given layer
        It is useful for debugging and reproducibility.
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        self._nb_params = sum([np.prod(p.size()) for p in model_parameters])
        return self._nb_params


class FastBatchNorm1d(nn.BatchNorm1d, BaseModule):
    def __init__(self, num_features, momentum=0.1):
        super(nn.BatchNorm1d, self).__init__(num_features, momentum=momentum)

    def _forward_dense(self, x):
        return super()(x)

    def _forward_sparse(self, x):
        """ Batch norm 1D is not optimised for 2D tensors. The first dimension is supposed to be
        the batch and therefore not very large. So we introduce a custom version that leverages BatchNorm1D
        in a more optimised way
        """
        x = x.unsqueeze(2)
        x = x.transpose(0, 2)
        x = super()(x)
        x = x.transpose(0, 2)
        return x.squeeze()

    def forward(self, x):
        if x.dim() == 2:
            return self._forward_sparse(x)
        elif x.dim() == 3:
            return self._forward_dense(x)
        else:
            raise ValueError("Non supported number of dimensions {}".format(x.dim()))


class _BNBase(nn.Sequential):
    def __init__(self, in_size, batch_norm=None, name=""):
        super(_BNBase, self).__init__()
        self.add_module(name + "bn", batch_norm(in_size))

        nn.init.constant_(self[0].weight, 1.0)
        nn.init.constant_(self[0].bias, 0)


class BatchNorm1d(_BNBase):
    def __init__(self, in_size, name=""):
        # type: (BatchNorm1d, int, AnyStr) -> None
        super(BatchNorm1d, self).__init__(in_size, batch_norm=nn.BatchNorm1d, name=name)


class BatchNorm2d(_BNBase):
    def __init__(self, in_size, name=""):
        # type: (BatchNorm2d, int, AnyStr) -> None
        super(BatchNorm2d, self).__init__(in_size, batch_norm=nn.BatchNorm2d, name=name)


class BatchNorm3d(_BNBase):
    def __init__(self, in_size, name=""):
        # type: (BatchNorm3d, int, AnyStr) -> None
        super(BatchNorm3d, self).__init__(in_size, batch_norm=nn.BatchNorm3d, name=name)


class _ConvBase(nn.Sequential):
    def __init__(
        self,
        in_size,
        out_size,
        kernel_size,
        stride,
        padding,
        dilation,
        activation,
        bn,
        init,
        conv=None,
        norm_layer=None,
        bias=True,
        preact=False,
        name="",
    ):
        super(_ConvBase, self).__init__()

        bias = bias and (not bn)
        conv_unit = conv(
            in_size,
            out_size,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )
        init(conv_unit.weight)
        if bias:
            nn.init.constant_(conv_unit.bias, 0)

        if bn:
            if not preact:
                bn_unit = norm_layer(out_size)
            else:
                bn_unit = norm_layer(in_size)

        if preact:
            if bn:
                self.add_module(name + "normlayer", bn_unit)

            if activation is not None:
                self.add_module(name + "activation", activation)

        self.add_module(name + "conv", conv_unit)

        if not preact:
            if bn:
                self.add_module(name + "normlayer", bn_unit)

            if activation is not None:
                self.add_module(name + "activation", activation)


class Conv1d(_ConvBase):
    def __init__(
        self,
        in_size,
        out_size,
        kernel_size=1,
        stride=1,
        padding=0,
        dilation=1,
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm1d,
    ):
        # type: (Conv1d, int, int, int, int, int, int, Any, bool, Any, bool, bool, AnyStr, _BNBase) -> None
        super(Conv1d, self).__init__(
            in_size,
            out_size,
            kernel_size,
            stride,
            padding,
            dilation,
            activation,
            bn,
            init,
            conv=nn.Conv1d,
            norm_layer=norm_layer,
            bias=bias,
            preact=preact,
            name=name,
        )


class Conv2d(_ConvBase):
    def __init__(
        self,
        in_size,
        out_size,
        kernel_size=(1, 1),
        stride=(1, 1),
        padding=(0, 0),
        dilation=(1, 1),
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm2d,
    ):
        # type: (Conv2d, int, int, Tuple[int, int], Tuple[int, int], Tuple[int, int], Tuple[int, int], Any, bool, Any, bool, bool, AnyStr, _BNBase) -> None
        super(Conv2d, self).__init__(
            in_size,
            out_size,
            kernel_size,
            stride,
            padding,
            dilation,
            activation,
            bn,
            init,
            conv=nn.Conv2d,
            norm_layer=norm_layer,
            bias=bias,
            preact=preact,
            name=name,
        )


class Conv3d(_ConvBase):
    def __init__(
        self,
        in_size,
        out_size,
        kernel_size=(1, 1, 1),
        stride=(1, 1, 1),
        padding=(0, 0, 0),
        dilation=(1, 1, 1),
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm3d,
    ):
        # type: (Conv3d, int, int, Tuple[int, int, int], Tuple[int, int, int], Tuple[int, int, int], Tuple[int, int, int], Any, bool, Any, bool, bool, AnyStr, _BNBase) -> None
        super(Conv3d, self).__init__(
            in_size,
            out_size,
            kernel_size,
            stride,
            padding,
            dilation,
            activation,
            bn,
            init,
            conv=nn.Conv3d,
            norm_layer=norm_layer,
            bias=bias,
            preact=preact,
            name=name,
        )


class FC(nn.Sequential):
    def __init__(
        self,
        in_size,
        out_size,
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=None,
        preact=False,
        name="",
    ):
        # type: (FC, int, int, Any, bool, Any, bool, AnyStr) -> None
        super(FC, self).__init__()

        fc = nn.Linear(in_size, out_size, bias=not bn)
        if init is not None:
            init(fc.weight)
        if not bn:
            nn.init.constant_(fc.bias, 0)

        if preact:
            if bn:
                self.add_module(name + "bn", BatchNorm1d(in_size))

            if activation is not None:
                self.add_module(name + "activation", activation)

        self.add_module(name + "fc", fc)

        if not preact:
            if bn:
                self.add_module(name + "bn", BatchNorm1d(out_size))

            if activation is not None:
                self.add_module(name + "activation", activation)



from jammy.utils.meter import GroupMeters
import wandb


class TrainerMonitor:
    def __init__(self, is_wandb=False, tb_logger=False):
        self.is_wandb = is_wandb
        self.counter = GroupMeters()
        if isinstance(tb_logger, bool):
            assert tb_logger == False
            self._tb_logger = None
            self.is_tblogger = False
        else:
            self._tb_logger = tb_logger
            self.is_tblogger = True

    def update(self, updates=None, value=None, **kwargs):
        """
        Example:
            >>> Monitor.update(key, value)
            >>> Monitor.update({key1: value1, key2: value2})
            >>> Monitor.update(key1=value1, key2=value2)
        """
        if updates is None:
            updates = {}
        if updates is not None and value is not None:
            updates = {updates: value}
        updates.update(kwargs)
        if self.is_wandb:
            wandb.log(updates)
        if self.is_tblogger:
            for k, v in updates.items():
                self.counter[k].update(1, 1)
                self._tb_logger.scalar_summary(k, v, self.counter[k].tot_count)

    # TODO: add a method for adding image here! 
    def flush(self):
        if self.is_wandb:
            # TODO:
            pass
        if self.is_tblogger:
            self._tb_logger.flush()

    def close(self):
        if self.is_wandb:
            wandb.finish()
        if self.is_tblogger:
            self._tb_logger.close()


#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : index.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 10/23/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.
from jamweb.web import JamRequestHandler, route


@route(r"/")
class IndexHandler(JamRequestHandler):  # pylint: disable=abstract-method
    def get(self):
        nr_visited = 0
        if self.session is not None:
            nr_visited = self.session.get("nr_visited", 0)
            self.session["nr_visited"] = nr_visited + 1
        self.render("index.html", nr_visited=nr_visited, headers=self.request.headers)



"""A CLI tool for sorting YAML files recursively.

This script provides functionality to load a YAML file, sort its contents
recursively, and save the sorted result to a new file. It can be used as a
command-line tool with various options for input, output, and sorting behavior.
"""

import argparse
import sys
from collections import OrderedDict
from typing import Any, Dict, List, Union

import yaml


def sort_dict(d: Dict[str, Any]) -> OrderedDict:
    """Sort a dictionary by its keys.

    Args:
        d: The dictionary to be sorted.

    Returns:
        An OrderedDict with sorted keys.
    """
    return OrderedDict(sorted(d.items(), key=lambda x: x[0]))


def dict_representer(dumper: yaml.Dumper, data: OrderedDict) -> yaml.nodes.MappingNode:
    """Custom representer for OrderedDict to maintain order in YAML output.

    Args:
        dumper: The YAML Dumper object.
        data: The OrderedDict to be represented.

    Returns:
        A YAML mapping node.
    """
    return dumper.represent_mapping("tag:yaml.org,2002:map", data.items())


yaml.add_representer(OrderedDict, dict_representer)


def sort_recursive(
    obj: Union[Dict[str, Any], List[Any], Any]
) -> Union[OrderedDict, List[Any], Any]:
    """Recursively sort a nested structure of dictionaries and lists.

    Args:
        obj: The object to be sorted. Can be a dict, list, or any other type.

    Returns:
        A sorted version of the input object.
    """
    if isinstance(obj, dict):
        return sort_dict({k: sort_recursive(v) for k, v in obj.items()})
    if isinstance(obj, list):
        return [sort_recursive(item) for item in obj]
    return obj


def load_yaml(file_path: str) -> Dict[str, Any]:
    """Load a YAML file and return its contents as a dictionary.

    Args:
        file_path: The path to the YAML file to be loaded.

    Returns:
        A dictionary containing the YAML file contents.

    Raises:
        yaml.YAMLError: If there's an error parsing the YAML file.
        FileNotFoundError: If the specified file is not found.
    """
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)
    except yaml.YAMLError as e:
        print(f"Error parsing YAML file: {e}", file=sys.stderr)
        sys.exit(1)
    except FileNotFoundError:
        print(f"File not found: {file_path}", file=sys.stderr)
        sys.exit(1)


def save_yaml(data: OrderedDict, file_path: str) -> None:
    """Save a dictionary as a YAML file.

    Args:
        data: The dictionary to be saved.
        file_path: The path where the YAML file should be saved.

    Raises:
        IOError: If there's an error writing to the file.
    """
    try:
        with open(file_path, "w", encoding="utf-8") as f:
            yaml.dump(data, f, default_flow_style=False)
    except IOError as e:
        print(f"Error writing to file: {e}", file=sys.stderr)
        sys.exit(1)


def main() -> None:
    """Main function to handle CLI arguments and execute the sorting process."""
    parser = argparse.ArgumentParser(description="Sort a YAML file recursively.")
    parser.add_argument("input", help="Input YAML file path")
    parser.add_argument(
        "-o", "--output", help="Output YAML file path (default: <input>_sorted.yaml)"
    )
    parser.add_argument(
        "--no-sort-lists", action="store_true", help="Don't sort list contents"
    )
    args = parser.parse_args()

    # Set default output file name if not provided
    if not args.output:
        args.output = args.input.rsplit(".", 1)[0] + "_sorted.yaml"

    # Load, sort, and save the YAML file
    config_dict = load_yaml(args.input)

    def sort_recursive_custom(
        obj: Union[Dict[str, Any], List[Any], Any]
    ) -> Union[OrderedDict, List[Any], Any]:
        if isinstance(obj, dict):
            return sort_dict({k: sort_recursive_custom(v) for k, v in obj.items()})
        if isinstance(obj, list) and not args.no_sort_lists:
            return [sort_recursive_custom(item) for item in obj]
        return obj

    sorted_config = sort_recursive_custom(config_dict)
    save_yaml(sorted_config, args.output)
    print(f"Sorted YAML saved to: {args.output}")


if __name__ == "__main__":
    main()
